/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is ServStats.
 *
 * The Initial Developer of the Original Code is
 * George J. Lee <gjl@csail.mit.edu>
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


// TODO: compare diagnosis with probe and without probe.  Figure out
// when to do a probe vs when not to do one... maybe just using a
// simple heuristic.  Well I'll leave it as is for now.  That way the
// learning is more accurate.

// TODO: add more ffox error codes cffus etc. to range of bnet.

// If you want to add a new property of an existing component, you
// need to make the following changes:
//
// 1. If you want it to be included in the depGraph sent to other
// agents, add the property info into the evidenceMap in
// insertEvidenceFromCIB().
//
// 2. Make sure you set the appropriate CIB entries using
// setCIBEntry() (e.g. in addFailureCIBEntries())
//
// 3. Modify serializeFailureObservation() to include the property you
// want sent.
//
// 4. Add your info to getFailureObservationsFromDepGraph().

const CLASS_ID = Components.ID("{d9143379-0241-4a39-a424-63ab66b60b57}");
const CLASS_NAME = "CAPRI Diagnosis User Agent Component";
const CONTRACT_ID = "@capri.csail.mit.edu/ua;1";
const DATA_ID = "kpd1";
const SRC_IP = '127.0.0.1'; // this gets set correctly at the regional agent
const STATE_START = Components.interfaces.nsIWebProgressListener.STATE_START;
const STATE_STOP  = Components.interfaces.nsIWebProgressListener.STATE_STOP;
const OK = 0;
const MAX_HIST_SIZE = 0; // send out all new connHist entries
const MAX_DIAG_HIST_SIZE = 0; // send out all new connHist entries
const CONNHIST_POLLING_PERIOD = 60 * 1000; // poll every minute

const ELAPSED_TIME_FAILURE_THRESH = 15 * 1000; // consider it a failure if
					       // a user hits stop or
					       // navigates away from a
					       // page after waiting 15
                                               // seconds.

const RETRY_TIMEOUT = 10 * 60 * 1000; // retry getting the history
				     // agent map every ten minutes

const CIB_LIFETIME = 60 * 60 * 1000 // keep CIB entries for one hour

const AGENT_MAP_LIFETIME = 24 * 60 * 60 * 1000; // get a new agent map
					       // every 24 hours.

const KNOWLEDGE_LIFETIME = 60 * 60 * 1000; // get new knowledge every hour.

const MIN_CONN_HIST_DELAY = 5000; // only add successive successes to
				  // the connHist if the time between
				  // them exceeds this. (5 seconds)

const DNS_TIMEOUT = 200 // only wait 200 ms for a dns lookup (i.e. just check the cache.)
const STATUS_INTERRUPTED = 2;
const NS_ERROR_UNEXPECTED = 0x8000ffff;

const NUM_HIST_AGENTS = 1; // only 1 history agent

const PROTO_VER = "0.5";
const CLIENT_AGENT = "ServStats";
const CLIENT_VER = "0.503";

const DIAG_DIR = "diag";
const CONN_HIST_FILE      = "conn-hist.xml";
const DIAG_HIST_FILE      = "diag-hist.xml";
const CONN_SUMMARY_FILE   = "conn-summary.rdf";
const AGENT_MAP_FILE      = "agent-map.xml";
const KNOWLEDGE_BASE_FILE = "kb.xml";
const PREFIX = "http://capri.csail.mit.edu/2006/capri/";

const CONN_SUMMARY_URI = PREFIX + "ua#connSummary";

const TEST_DIR = '/home/gjl/capri/mozdev/servstats/src/test/';

const RDFC_ID = '@mozilla.org/rdf/container;1';
const RDFCUtils_ID = '@mozilla.org/rdf/container-utils;1';
var RDFCUtils = Components.classes[RDFCUtils_ID].getService(Components.interfaces.nsIRDFContainerUtils);

const RDF_ID = '@mozilla.org/rdf/rdf-service;1';
var RDF = Components.classes[RDF_ID].getService(Components.interfaces.nsIRDFService);
/*****************************************************************************
 * diag component
 *****************************************************************************/
function diagAgent() {};
diagAgent.prototype = {
  initialized: false,
  pref: null, // XPCOM preferences
  obs: null, // the XPCOM observer service
  connHist: null,  // an associative array of arrays of connection entries
  connSummary: null,
  debug: false,
  alertId: 0,
  reqId: 1,  // initialize the request ID to 1

  diagHist: [], // record of previous local diagnoses
  diagHistEntriesSent: 0, // how entries sent to the regional agent

  consecFailedDests: {}, // the number of consecutive failed servers

  // confidence threshold of 95%
  confThresh: 0.95,
  knowledgeBase: {}, // probabilistic dependency knowledge
  componentInformationBase: {}, // cache of information about components

  componentId: 1, // initialize component ID to 1

  lastFailure: null, // the entry for the most recent failure
  
  // Keep track of the start time of each active connection to detect
  // unresponsive servers.
  startTime: null,

  // each diagHistory entry contains these fields (and their
  // corresponding Time fields)
  diagAttributes: ['destHash', 'cffu', 'cfts', 'probeResult',
		   'probeURI', 'ffoxErrorCode', 'diagResult'],

  // knowledge about classes of components, including dependencies,
  // range, parents, and children.
  nodeAttributes: {
    cause: {
      range: ['lwdi', 'lwdI', 'lwDi', 'lwDI',
	      'lWdi', 'lWdI', 'lWDi', 'lWDI',
	      'Lwdi', 'LwdI', 'LwDi', 'LwDI',
	      'LWdi', 'LWdI', 'LWDi', 'LWDI'],
      parents: [],
      children: ['obconn', 'cffu', 'http', 'cfts', 'ffox'],
      cpt: {},
    },
    obconn: {
      range: ['OK', 'FAIL'],
      parents: ['cause'],
      children: [],
      cpt: {},
    },
    cffu: {
      range: [0, 1, 2, 3],
      parents: ['cause'],
      children: [],
      cpt: {},
    },
    http: {
      range: ['OK', 'FAIL'],
      parents: ['cause'],
      children: [],
      cpt: {},
    },
    cfts: {
      range: [0, 1, 2, 3],
      parents: ['cause'],
      children: [],
      cpt: {},
    },
    ffox: {
      range: [0, 2, 13, 14, 20, 21, 22, 30, 42, 71, 72],
      parents: ['cause'],
      children: [],
      cpt: {},
    },
  },

  getPropertyAbbrev: function() {
    // These property names are only used in the knowledge base
    abbrevs = {};
    abbrevs[PREFIX + 'common#Local_Network' + ' ' +
	    PREFIX + 'core#status']                  = 'localnet.status';
    abbrevs[PREFIX + 'common#Local_Network' + ' ' +
	    PREFIX + 'common#consecFailuresFromUser']   = 'localnet.cffu';
    abbrevs[PREFIX + 'common#HTTP_Connection' + ' ' +
	    PREFIX + 'core#status']                  = 'http.status';
    abbrevs[PREFIX + 'common#DNS_Lookup' + ' ' +
	    PREFIX + 'core#status']                  = 'dns.status';
    abbrevs[PREFIX + 'common#IP_Routing' + ' ' +
	    PREFIX + 'core#status']                  = 'ip.status';
    abbrevs[PREFIX + 'common#AS_Hop' + ' ' +
	    PREFIX + 'core#status']                  = 'ashop.status';
    abbrevs[PREFIX + 'common#HTTP_Server' + ' ' +
	    PREFIX + 'core#status']                  = 'webserver.status';
    abbrevs[PREFIX + 'common#HTTP_Server' + ' ' +
	    PREFIX + 'common#consecFailuresToServer']   = 'webserver.cfts';
    abbrevs[PREFIX + 'common#Verify_DNS_Lookup_Test' + ' ' +
	    PREFIX + 'common#dnsLookupResult'] = 'verifydns.result';
    abbrevs[PREFIX + 'common#Outbound_Conn_Test' + ' ' +
	    PREFIX + 'common#probeResult']               = 'obconn.result';
    abbrevs[PREFIX + 'common#Firefox_Error_Test' + ' ' +
	    PREFIX + 'common#ffoxErrorCode']            = 'ffox.error';
    abbrevs[PREFIX + 'common#AS_Hop_Test' + ' ' +
	    PREFIX + 'common#asHopTestResult']         = 'ashoptest.result';
    return abbrevs;
  },

  alert: function(msg) {
    this.alertId++;
    var win = this.getWindow();
    var response = "chrome://diag/content/alert.xul";
    win.openDialog(response, "Alert " + this.alertId,
		   "chrome,all,dialog=no", msg);
  },

  dump: function(aMessage) {
    if (this.debug) {
      var consoleService = Components.classes["@mozilla.org/consoleservice;1"].getService(Components.interfaces.nsIConsoleService);
      consoleService.logStringMessage("DiagAgent: " + aMessage);
    }
  },

  getContainer: function(ds, rootnode) {
    var con = Components.classes[RDFC_ID].createInstance(Components.interfaces.nsIRDFContainer);
    con.Init(ds, rootnode);
    return con;
  },

  getProfileFile: function(filename) {
    // get profile directory
    var file = Components.classes["@mozilla.org/file/directory_service;1"]
    .getService(Components.interfaces.nsIProperties)
    .get("ProfD", Components.interfaces.nsIFile);

    file.append(DIAG_DIR);

    // if it doesn't exist, create
    if( !file.exists() ) {   
      file.create(Components.interfaces.nsIFile.DIRECTORY_TYPE, 0755);
    }

    file.append(filename);
    return file;
  },

  createDatasource: function(filename, uri, reset) {
    var ds;
    var file = this.getProfileFile(filename);
    var fileURI = this.getFileURI(file);

    // if it doesn't exist, create
    if( reset || !file.exists() ) {
      this.dump("Creating new datasource");
      ds = Components.classes["@mozilla.org/rdf/datasource;1?name=xml-datasource"].createInstance(Components.interfaces.nsIRDFDataSource);

      var rootnode = RDF.GetResource(uri);
      RDFCUtils.MakeSeq(ds, rootnode);
      var con = this.getContainer(ds, rootnode);
      ds = ds.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource);
      ds.FlushTo(fileURI);
    } else {
      ds = RDF.GetDataSourceBlocking(fileURI);
    }
    return ds;
  },
  
  readDocument: function(aFile)
  // return a doc containing the contents of the XML file with name aFile
  {
    //setup input stream    
    var contents = new String();    
    var fiStream = Components.classes["@mozilla.org/network/file-input-stream;1"].createInstance(Components.interfaces.nsIFileInputStream);
    var siStream = Components.classes["@mozilla.org/scriptableinputstream;1"].createInstance(Components.interfaces.nsIScriptableInputStream);         
    var converter = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].createInstance(Components.interfaces.nsIScriptableUnicodeConverter);    
    converter.charset = "UTF-8";    

    try {
      //load file into stram
      fiStream.init(aFile, 1, 0, false);
      siStream.init(fiStream);
      while (siStream.available() > 0) {
	var chunk = siStream.read(siStream.available());
	contents += converter.ConvertToUnicode(chunk);      
      };
      
      //close file
      siStream.close();                          
      fiStream.close();
    } catch (e) {
      this.dump("Error reading file " + aFile);
      return null;
    }

    var parser = Components.classes["@mozilla.org/xmlextras/domparser;1"].createInstance(Components.interfaces.nsIDOMParser);           
    var doc = parser.parseFromString(contents, "text/xml");

    return doc;
  },
  
  loadAgentMap: function(reload) {
    // If reload is true, then get a new copy of the agent map from
    // the directory server.  Otherwise, it's permissible to read the
    // agent map from disk.  This gets called at startup and when the
    // user changes the options.  The only time we should read from
    // disk is on startup.

    var agentMapFile = this.getProfileFile(AGENT_MAP_FILE);
    var doc = this.readDocument(agentMapFile);

    // if our map is too old, get a new one.
    var agentMapExpires = 1000 * this.prefs.getIntPref("extensions.diagtool.agentMapExpires");
    var curTime = this.getCurrentTime();

    if (reload || doc == null || agentMapExpires < curTime) {
      this.requestAgentMap(reload);
    } else {
      this.agentMapDoc = doc;
      this.parseAgentMap(doc);
    }
  },

  loadConnHist: function()
  // Read in the connHist file into an associative array
  {
    var connHistFile  = this.getProfileFile(CONN_HIST_FILE);
    var doc = this.readDocument(connHistFile);

    var connHist = {entries: [], entriesSent: 0};

    if (doc == null) {
      // i/o error or no file
      return connHist;
    }

    var connHistQueue = connHist.entries;
    var rootNode = doc.documentElement;
    var hosts = rootNode.getElementsByTagName("host");
    for (var i=0; i<hosts.length; i++) {
      var host = hosts[i];
      var entry = {};
      var hostHash = hosts[i].getAttribute("hash");
      entry.hostHash = hostHash;
      
      var conns = host.getElementsByTagName('conn');
      for (var j=0; j<conns.length; j++) {
        entry.connTime = conns[j].getAttribute("ct");
        entry.elapsedTime = conns[j].getAttribute("et");
        entry.status = conns[j].getAttribute("st");
        entry.errorCode = conns[j].getAttribute("ec");
	if (conns[j].hasAttribute("ip")) {
	  entry.destIP = conns[j].getAttribute("ip");
	}
	  
      }
      connHistQueue.push(entry);
    }
    return connHist;
  },

  loadKnowledgeBase: function()
  {
    this.dump("Loading knowledge base");
    var knowledgeBaseFile = this.getProfileFile(KNOWLEDGE_BASE_FILE);
    var doc = this.readDocument(knowledgeBaseFile);

    // if our knowledge is too old, get new knowledge from the
    // regional agent.
    var knowledgeExpires = 1000 * this.prefs.getIntPref("extensions.diagtool.knowledgeExpires");
    var curTime = this.getCurrentTime();

    if (doc == null || knowledgeExpires < curTime) {
      // file doesn't exist or i/o error, so request it from the
      // regional agent.
      this.sendKnowledgeRequest();
      
      this.dump("loadKnowledgeBase: requesting new knowledge base (doc == " +
		doc + ", knowledgeExpires == " + knowledgeExpires +
		", curTime == " + curTime);
      return;
    }

    this.parseKnowledge(doc);
  },

  loadDiagHist: function()
  {
    // Load a file of diagnosis histories into an array diagHist
    //

    var diagHistFile = this.getProfileFile(DIAG_HIST_FILE);
    var doc = this.readDocument(diagHistFile);

    var diagHist = [];

    if (doc != null) {
      var diags = doc.getElementsByTagName("diag");
      for (var i=0; i<diags.length; i++) {
	var diagAttributes = this.diagAttributes;

	var entry = {}
	for (var diagAttrId=0; diagAttrId<diagAttributes.length; diagAttrId++) {
	  var attribute = diagAttributes[diagAttrId];
	  entry[attribute] = {};
	  entry[attribute].value = diags[i].getAttribute(attribute);
	  entry[attribute].time  = diags[i].getAttribute(attribute + 'Time');
	}
      
	if (entry.destHash.value == null || entry.destHash.time == null) {
	  this.dump("Malformed diagHist entry");
	} else {
	  diagHist.push(entry);
	}
      }
    }

    return diagHist;
    
  },    

  trim: function(s) {
    // strip leading and trailing whitespace from a string
    return s.replace(/\s+/g, '');
  },

  parseObservations: function(doc)
  {
    // Extract observations from an XML message.
    
    // get all observations
    var obsList = [];
    var observations = doc.getElementsByTagName("observation");
    for (var i=0; i<observations.length; i++) {
      // component node is the child of an observation node
      var obsNode = observations[i];
      var comNode = null;
      for (var childIdx=0; childIdx < obsNode.childNodes.length; childIdx++) {
	if (obsNode.childNodes[childIdx].nodeType == obsNode.ELEMENT_NODE) {
	  comNode = obsNode.childNodes[childIdx];
	}
      }
      if (comNode == null) {
	this.dump('parseObservations(): Error, could not get com node');
	continue;
      }
      // for each observation, get its hosthash and isdead property
      var classNodes = comNode.getElementsByTagName("class");
      if (classNodes.length != 1) {
	this.dump("Warning, found " + classNodes.length + " classNodes");
	continue;
      }
      var className = classNodes[0].textContent;

      if (className == null) {
	this.dump("className == null");
	continue;
      }
	
      // strip whitespace from class URI
      className = className.replace(/\s+/g, '');
	
      var obs = {
	'className': className,
	'props': {},
	'metadata': {},
      };
      for (var childIdx=0; childIdx < comNode.childNodes.length; childIdx++) {
	var propNode = comNode.childNodes[childIdx];
	if (propNode.nodeType != comNode.ELEMENT_NODE) {
	  continue;
	}
	if (propNode.tagName == 'class') {
	  continue;
	}
	var namespaceURI = '';
	if (propNode.namespaceURI != null) {
	  namespaceURI = propNode.namespaceURI;
	}
	var propName = namespaceURI + propNode.localName;
	var propVal = propNode.textContent;
	// this.dump('propName: ' + propName + ', propVal: ' + propVal);
	var time = propNode.getAttribute('time');
	obs.props[propName] = propVal;
	obs.metadata[propName] = {'time': time};
      }

      if (obs.className == PREFIX + 'common#HTTP_Server') {
	var cftsPropName = PREFIX + 'common#consecFailuresToServer';
	var cfts = obs.props[cftsPropName];
	if (cfts != undefined) {
	  cfts = parseInt(cfts);
	  var hostHash = obs.props[PREFIX + 'common#hostHash'];
	  var time = obs.metadata[cftsPropName]['time'];
	  this.setCIBEntry(time, className, hostHash,
			   PREFIX + 'common#consecFailuresToServer', cfts);
	  this.dump("cib[" + obs.className + "][" + hostHash +
		    "]['" + cftsPropName + "'] = " + cfts);
	}
      }

      obsList.push(obs);
    }
    return obsList;
  },

  requestServStatsTest: function() {
    this.dump("Testing requestServStats():");
    this.dump("Testing null destination");
    this.currentHost = 'www.google.com';
    this.requestServStats();
    // should test www.google.com
    
    this.dump("Testing localhost destination");
    this.requestServStats('localhost');
    // should pop up an alert saying server stats not available for localhost

    this.dump("Testing remote destination");
    this.requestServStats('capri.csail.mit.edu');
    // should pop up servstats for capri.csail.mit.edu
  },

  requestDiagnosisTest: function() {
    this.dump("Testing requestDiagnosis():");

    this.lastFailure = null;
    this.dump("Testing no recent failures:");
    this.requestDiagnosis();
    // should return an alert: "No recent failures to diagnose."

    this.lastFailure = {host: "localhost", connTime: 1234, status: 14};
    this.dump("Testing null destination:");
    this.requestDiagnosis();
    // should return an alert: "Sorry, diagnosis is not available for
    // localhost."

    this.dump("Testing localhost destination:");
    this.requestDiagnosis('localhost');
    // should return an alert: "Sorry, diagnosis is not available for
    // localhost."

    var oldConfThresh = this.confThresh;

    // Simulate a failure of capri.csail.mit.edu
    

    this.confThresh = 0;
    this.dump("Testing local diagnosis (confThresh = 0):");
    this.requestDiagnosis('capri.csail.mit.edu');
    // should return a local diagnosis.

    this.confThresh = 0.8;
    this.dump("Testing remote diagnosis (confThresh = 0.8):");
    this.requestDiagnosis('capri.csail.mit.edu');

    this.confThresh = oldConfThresh;
    // should do a remote diagnosis.
  },

  serializeKnowledgeBaseTest: function() {
    // assume we have already parsed knowledge
    this.writeKnowledgeBase();
  },

  createNotificationTest: function() {
    var curTime = this.getCurrentTime();
    var hostname = 'foo';
    var entry = {host: hostname,
	     hostHash: this.md5sum(hostname),
	     connTime: curTime,
	     errorCode: 30,
	     status: 'FAIL'
    };
    
    this.updateConnHist(entry);

    entry = {host: hostname,
	     hostHash: this.md5sum(hostname),
	     connTime: curTime,
	     errorCode: 0,
	     status: 'OK'
    };
    
    this.updateConnHist(entry);

    var notifications = this.connHistToNotifications();

    this.dump('History notification: ' + notifications.join('\n'));
  },

  diagHistToNotificationsTest: function() {
    // note that this test wipes out your diaghist
    var entry = {}
    var diagAttributes = this.diagAttributes;
    
    // Insert a bogus entry
    for (var diagAttrId=0; diagAttrId<diagAttributes.length; diagAttrId++) {
      var attribute = diagAttributes[diagAttrId];
      entry[attribute] = {};
      entry[attribute].value = attribute;
      entry[attribute].time  = attribute + 'Time';
    }
    this.diagHist = [entry];
    var notifications = this.diagHistToNotifications();
    this.dump('diagHistNotificationTest: ' + notifications.join('\n'));
  },

  dnsLookupAndProcessEntry: function(entry) {
    // Given a failure entry, do a DNS lookup and set the ipAddrs and
    // destIP fields of the entry.  Use asynchronous version to avoid
    // locking up the browser.
    //
    // Watch out for possible race condition.  This might cause two
    // diagnoses of the same failure.
    //

    var agent = this;
    var dnsTimer = Components.classes["@mozilla.org/timer;1"].createInstance(Components.interfaces.nsITimer);

    var dnsHandle = null;

    var dnsTimerCallback = {
      notify: function(timer) {
	if (dnsHandle != null) {
	  agent.dump("dnsTimerCallback: Canceling DNS lookup for " +
		     entry.host);
	  // Canceling the lookup causes dnsHandler.onLookupComplete
	  // to be called
	  dnsHandle.cancel(NS_ERROR_UNEXPECTED);
	} else {
	  agent.dump("Warning, dnsHandle is null for " + entry.host);
	}
	agent.processEntry(entry);
      }
    };

    // cancel the lookup after DNS_TIMEOUT milliseconds to ensure
    // prompt diagnosis.
    dnsTimer.initWithCallback(dnsTimerCallback, DNS_TIMEOUT,
			      Components.interfaces.nsITimer.TYPE_ONE_SHOT);

    var dnsHandler = {
      onLookupComplete: function(request, nsRecord, status) {
	// dnsTimer.cancel();
	var ipAddrs = [];

	agent.dump("DNS lookup for " + entry.host + " complete.  Status: " +
		   status);
 	while (nsRecord != null && nsRecord.hasMore()) {
	  var ipAddr = nsRecord.getNextAddrAsString();
	  ipAddrs.push(ipAddr);
	}

	if (ipAddrs.length > 0) {
	  // use the first IP address as the destIP
	  var destIP = ipAddrs[0];
	  agent.dump("Setting destIP = " + destIP + " for " + entry.host);
	  entry.destIP = destIP;
	} else {
	  agent.dump("No IP addresses found for " + entry.host);
	}

	// List ip addresses in order for consistency and caching
	ipAddrs.sort()
	entry.ipAddrs = ipAddrs.join(',');

	// Note, probehosts fails if we do agent.processEntry within
	// the DNS lookup callback function.  Therefore do it in the
	// timer callback.
      }
    };

    // dnsHandler.onLookupComplete(null, null, 0);
    dnsHandle = this.dnsService.asyncResolve(entry.host, 0,
					     dnsHandler, null);

  },

  processEntry: function(entry) {
    // After a page finishes loading, log it and diagnose it if
    // necessary.
    this.logConnection(entry);

    if (entry.status == 'FAIL') {
      this.dump("**** FAILURE. Host: " + entry.host +
		", error code: " + entry.errorCode);
      this.setLastFailure(entry);
      this.diagnoseFailure(entry);
    }
  },

  dnsLookupTest: function() {
    entry = {host: 'google.com', status: 'OK'}
    this.dnsLookupAndProcessEntry(entry);
  },

  runTests: function() {
    this.dnsLookupTest();
    this.createNotificationTest();
    this.parseObservationsTest();
    this.parseKnowledgeTest();
    this.serializeKnowledgeBaseTest();

    // This is a naughty test because it corrupts diagHist
    // this.createDiagHistNotificationTest();

    this.inferenceTest();
    this.requestServStatsTest();

    // must insert the appropriate CIB entries before doing the
    // requestdiagnosistest
    this.requestDiagnosisTest();
  },

  parseObservationsTest: function()
  {
    var observationTestFile = this.getProfileFile('observationTest.xml');
    var doc = this.readDocument(observationTestFile);
    this.parseObservations(doc);

    var className = PREFIX + 'common#HTTP_Server';
    var hostHash = 'foo';
    var cfts = this.getCIBEntry(className,
				hostHash,
				PREFIX + 'common#consecFailuresToServer').value;
    if (cfts == 2) {
      this.dump("parseObservationsTest() passed.");
    } else {
      this.dump("parseObservationsTest() FAILED: cfts == " + cfts);
    }
  },
  
  parseKnowledgeTest: function()
  {
    var knowledgeTestFile = this.getProfileFile('knowledgeTest.xml');
    var doc = this.readDocument(knowledgeTestFile);
    this.parseKnowledge(doc);
    
    if (this.knowledgeBase['ffox.error'].cpt['30|OK OK FAIL FAIL'] == 0.81) {
      this.dump("parseKnowledgeTest() passed.");
    } else {
      this.dump("parseKnowledgeTest() FAILED.");
    }
  },

  parseKnowledge: function(doc)
  {
    // Extract probabilistic dependency knowledge from an XML message
    // doc and store it in a knowledge base
    //
    //
    this.dump("parseKnowledge(): starting");
    var knowledge = doc.getElementsByTagName("knowledge");
    var numEntries = 0;
    for (var i=0; i<knowledge.length; i++) {
      // Identify the component and parents
      var subjectNode = knowledge[i].getElementsByTagName("subject");
      var subjectURI = null;

      if (subjectNode.length != 1) {
	this.dump("Expected subjectNode.length == 1; " +
		  "subjectNode.length == " + subjectNode.length);
      } else {
	subjectURI = this.trim(subjectNode[0].textContent);
      }

      var propertyNode = knowledge[i].getElementsByTagName("property");
      var propertyURI = null;
      
      if (propertyNode.length != 1) {
	this.dump("Expected propertyNode.length == 1; " +
		  "propertyNode.length == " + propertyNode.length);
      } else {
	propertyURI = this.trim(propertyNode[0].textContent);
      }
	
      // abbreviate the subject and property URI in our internal
      // representation:
      var propertyName = this.propertyAbbrev[subjectURI + " " + propertyURI];
      if (propertyName == undefined) {
	this.dump("Unknown subjectURI and propertyURI " +
		  subjectURI + " " + propertyURI);
      }

      // overwrite any previous knowledge (may not be the best thing
      // to do in general)
      // this.knowledgeBase[subjectURI] = {};
      
      // Read in the conditional probability table
      var cptNode = knowledge[i].getElementsByTagName("cpt");
      if (cptNode.length != 1) {
	this.dump("Expected one CPT, but cptNode.length == " + cptNode.length);
      } else {
	var parentsNode = cptNode[0].getElementsByTagName("parents");
	if (parentsNode == null || parentsNode.length != 1) {
	  this.dump("Expected one parents, but parents.length==" + parentsNode.length);
	} else {
	  // Identify parents
	  var parentText = parentsNode[0].textContent;
	  var parents = '';
	  if (parentText != null) {
	    // parents not actually used below...
	    parents = parentText.split(/\s+/);
	  }
	  var entryNodes = cptNode[0].getElementsByTagName("entry");
	  for (var entryIndex = 0; entryIndex < entryNodes.length; entryIndex++) {
	    var entry = entryNodes[entryIndex];
	    var value = entry.getAttribute("value");
	    var parentVals = entry.getAttribute("parentVals");
	    if (parentVals == null) {
	      parentVals = '';
	    }
	    parentVals = parentVals.replace(/\s+/g, " ");
	    var prob = entry.getAttribute("p");
	    var key = value + "|" + parentVals;
	    
	    // Store the CPT in the knowledge base, indexed by
	    // component class.  Not the ideal representation...
	    if (this.knowledgeBase[propertyName] == undefined) {
	      this.knowledgeBase[propertyName] = {subject: subjectURI,
						  property: propertyURI,
						  node: knowledge[i],
						  
						  cpt: {}};
	    }
	    this.knowledgeBase[propertyName].cpt[key] = parseFloat(prob);
	    numEntries++;
 	    // this.dump("Inserting CPT entry: knowledgeBase['" + propertyName +
	    //      "'].cpt['" + key + "'] = " + prob);
	  }
	}
      }
    }
    this.dump("parseKnowledge(): inserted " + numEntries + " entries.");
    // translate and store the knowledge in the nodeattributes table
    if (knowledge.length > 0) {
      this.translateKnowledge();
    }
  },

  translateKnowledge: function() {
    // Take the knowledge in the knowledge base and translate it into
    // a form suitable for use as CPTs for node attributes in the
    // failure dependency graph.
    //
    // Precondition: the CPTs in the knowledge base are completely
    // filled in.

    // HTTP:
    // translate the parent values into the abbreviated form.
    // Parents listed in following order: Local net, Web server,
    // DNS, IP
    //
    if (!('http.status' in this.knowledgeBase)) {
      this.dump("Don't have enough knowledge to translate.");
      return;
    }
    var cpt = this.knowledgeBase['http.status'].cpt;
    for (var key in cpt) {
      var keyPieces = key.split('|');
      var value = keyPieces[0];
      var parents = keyPieces[1];
      var parentList = parents.split(' ');
      var base = 'lwdi';
      var newParents = '';
      for (var i=0; i<parents.length; i++) {
	var c;
	if (parentList[i] == 'OK') {
	  c = base.charAt(i).toUpperCase();
	} else {
	  c = base.charAt(i).toLowerCase();
	}
	newParents += c;
      }
      var newKey = value + '|' + newParents;
      var prob = cpt[key];
      this.nodeAttributes['http'].cpt[newKey] = prob;
//       this.dump("nodeAttributes['http'].cpt[" + newKey + "] = " + prob);
    }

    // Local net, IP, DNS, and IP routing:
    // compute the joint prior probabilities of all causes
    var causeRange = this.nodeAttributes['cause'].range;
    var names = ['localnet.status', 'webserver.status',
		 'dns.status', 'ip.status'];
    var sum = 0; // sanity check
    for (var i=0; i<causeRange.length; i++) {
      var key;
      var prob = 1.0;
      var causeValue = causeRange[i];
      for (var j=0; j<names.length; j++) {
	var statusChar = causeValue[j];
	if (statusChar == statusChar.toUpperCase()) {
	  // capital letter
	  key = 'OK|';
	} else {
	  key = 'FAIL|';
	}
	var priorProb = this.knowledgeBase[names[j]].cpt[key];
	if (priorProb == undefined) {
	  this.dump("key " + key + " undefined for " + names[j] + "!");
	}
	prob *= priorProb;
      }
      var newKey = causeValue + '|';
      this.nodeAttributes['cause'].cpt[newKey] = prob;
//       this.dump("nodeAttributes['cause'].cpt[" +
// 		    newKey + "] = " + prob);
      sum += prob;
    }

    if (Math.round(sum * 10000) != 10000) {
      this.dump("Sum is incorrect: " + sum);
    }

    // obconn: expand the CPT from a obconn.result|localnet.status
    // into a obconn.result|lwdi
    //
    // cffu: expand the CPT from a localnet.cffu|localnet.status
    // into a localnet.cffu|lwdi
    //
    // cfts: expand the CPT from webserver.cfts|webserver.status
    // into a webserver.cfts|lwdi
    //
    for (var i=0; i<causeRange.length; i++) {
      var prob;
      var parents;
      var causeValue = causeRange[i];
      var children = ['obconn', 'cffu', 'cfts'];
      var propertyNames = ['obconn.result', 'localnet.cffu', 'webserver.cfts'];
      var parentIndex = [0, 0, 1];
      for (var childId=0; childId<children.length; childId++) {
	var child = children[childId];
	var childRange = this.nodeAttributes[child].range;
	
	for (var j=0; j<childRange.length; j++) {
	  var childValue = childRange[j];
	  var statusChar = causeValue.charAt(parentIndex[childId]);
	  if (statusChar == statusChar.toUpperCase()) {
	    parents = 'OK';
	  } else {
	    parents = 'FAIL';
	  }
	  var key = childValue + '|' + parents;
	  var propertyName = propertyNames[childId];
	  prob = this.knowledgeBase[propertyName].cpt[key];
	  var newKey = childValue + '|' + causeValue;
	  this.nodeAttributes[child].cpt[newKey] = prob;
// 	  this.dump("nodeAttributes['" + child + "'].cpt[" +
// 		    newKey + "] = " + prob);
	}
      }
    }

    // ffox: convert the CPT from
    // ffox.error|localnet.status,webserver.status,dns.status,ip.status
    // to ffox.error|lwdi
    var numParents = 4;
    for (var i=0; i<causeRange.length; i++) {
      var key;
      var prob = 1.0;
      var causeValue = causeRange[i];
      
      var ffoxRange = this.nodeAttributes['ffox'].range;
      
      // For each ffox error code value
      for (var ffoxValueId = 0; ffoxValueId<ffoxRange.length; ffoxValueId++) {
	var ffoxValue = ffoxRange[ffoxValueId];

	// build up the list of parent values
	var parentVals = new Array();
	for (var j=0; j<numParents; j++) {
	  var statusChar = causeValue.charAt(j);
	  if (statusChar == statusChar.toUpperCase()) {
	    // capital letter
	    parentVals.push('OK');
	  } else {
	    parentVals.push('FAIL');
	  }
	}
	var parents = parentVals.join(' ');
	var key = ffoxValue + '|' + parents;
	var newKey = ffoxValue + '|' + causeValue;
	var prob = this.knowledgeBase['ffox.error'].cpt[key];
	this.nodeAttributes['ffox'].cpt[newKey] = prob;
// 	this.dump("nodeAttributes['ffox'].cpt[" + newKey + "] = " + prob);
      }
    }
  },

  writeFile: function(aFile, aText) {
    // write the string aText to a file aFile
    
    //setup output stream
    var foStream = Components.classes["@mozilla.org/network/file-output-stream;1"].createInstance(Components.interfaces.nsIFileOutputStream);     
    var converter = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].createInstance(Components.interfaces.nsIScriptableUnicodeConverter);             
    converter.charset = "UTF-8";    
    
    //initialize file stream
    var flags = 0x02 | 0x08 | 0x20;
    foStream.init(aFile, flags, 0664, 0);  
         
    //use string converter
    var chunk = converter.ConvertFromUnicode(aText);
    foStream.write(chunk, chunk.length);
    var fin = converter.Finish();
    if (fin.length > 0)
      foStream.write(fin, fin.length);
    
    //close file stream
    foStream.close(); 
  },

  writeAgentMap: function()
  {
    if (this.agentMapDoc != null) {
      this.dump('Writing agent map');
      var file = this.getProfileFile(AGENT_MAP_FILE);
      var xml = this.serializeXMLDoc(this.agentMapDoc);
      this.writeFile(file, xml);
    }
  },

  writeConnSummary: function()
  {
    this.dump("Writing connSummary");
    var connSummaryFile = this.getProfileFile(CONN_SUMMARY_FILE);
    this.connSummary = this.connSummary.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource);
    var fileURI = this.getFileURI(connSummaryFile);
    this.connSummary.FlushTo(fileURI);
  },

  writeConnHist: function()
  {
    this.dump("Writing connHist");
    var file = this.getProfileFile(CONN_HIST_FILE);
    var xml = this.serializeConnHist();
    this.writeFile(file, xml);
  },

  writeKnowledgeBase: function()
  {
    // write a file of dead hosts to disk
    this.dump("Writing knowledge base");
    var file = this.getProfileFile(KNOWLEDGE_BASE_FILE);
    var xml = this.serializeKnowledgeBase();
    this.writeFile(file, xml);
  },

  writeDiagHist: function()
  {
    // write a diag history to disk
    this.dump("Writing diagHist");
    var file = this.getProfileFile(DIAG_HIST_FILE);
    var xml = this.serializeDiagHist();
    this.writeFile(file, xml);
  },

  startTimer: function()
  {
    this.timer = Components.classes["@mozilla.org/timer;1"].createInstance(Components.interfaces.nsITimer);
    this.timer.init(this, CONNHIST_POLLING_PERIOD,
		    Components.interfaces.nsITimer.TYPE_REPEATING_SLACK);
  },

  stopTimer: function()
  {
    this.timer.cancel();
  },

  addObservers: function()
  {
    var obs = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
    obs.addObserver(this, "http-on-examine-response", true);
    obs.addObserver(this, "profile-before-change", true);
    obs.addObserver(this, "timer-callback", true);
    this.obs = obs;
  },

  removeObservers: function()
  {
      this.obs.removeObserver(this, "http-on-examine-response");
      this.obs.removeObserver(this, "profile-before-change");
      this.obs.removeObserver(this, "timer-callback");
  },

  updateVersion: function() {
    // If this is the first time running a new version, upgrade any
    // necessary files.

    var previousVersion = null;
    try {
      previousVersion = this.prefs.getCharPref("extensions.diagtool.version");
    } catch (e) {
      this.dump('updateVersion: exception caught: ' + e);
    }
    
    if (previousVersion != CLIENT_VER) {
      this.dump('Upgrading from version ' + previousVersion +
		' to ' + CLIENT_VER);
      // delete old obsolete files
      try {
	var file;
	
	file = this.getProfileFile(CONN_SUMMARY_FILE);
	if (file != null) {
	  file.remove(false);
	}
	
	file = this.getProfileFile(CONN_HIST_FILE);
	if (file != null) {
	  file.remove(false);
	}
      } catch (e) {
	this.dump(e);
      }
    }

    this.prefs.setCharPref("extensions.diagtool.version", CLIENT_VER);
  },
  
  start: function()
  {
    // do nothing if already loaded
    if (this.initialized) {
      return;
    }

    // retrieve preferences
    this.prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);
    this.debug = this.prefs.getBoolPref('extensions.diagtool.debug');
    this.dump("Starting diagAgent");

    // If this is the first time running a new version, upgrade any
    // necessary files.
    this.updateVersion();

    this.sendHostname = this.prefs.getBoolPref("extensions.diagtool.sendHostname");
    try {
      var dnsClass = Components.classes['@mozilla.org/network/dns-service;1'];
      var dnsIFace = Components.interfaces.nsIDNSService;
      this.dnsService = dnsClass.getService(dnsIFace);
    } catch (e) {
      this.dump(e)
    }

    // create RDF datasources
    this.connSummary = this.createDatasource(CONN_SUMMARY_FILE,
					     CONN_SUMMARY_URI);

    this.propertyAbbrev  = this.getPropertyAbbrev();
    this.connHist        = this.loadConnHist();
    this.diagHist        = this.loadDiagHist();

    if (this.prefs.getBoolPref("extensions.diagtool.autoLookup")) {
      this.loadAgentMap();
    }
    this.loadKnowledgeBase();

    // add observers
    this.addObservers();

    // to periodically poll connHist
    this.startTimer();

    this.initialized = true;
    this.startTime = {};

    this.agentMapDoc = null;
    this.lastFailure = null;
    this.currentHost = null;
  },
  
  stop: function()
  {
    // do nothing if not loaded
    if (!this.initialized) {
      return;
    }
    
    // clean up held observers etc to avoid leaks.
    try {
      this.writeConnHist();
      this.writeConnSummary();
      this.writeDiagHist();
      this.writeAgentMap();
      
      this.stopTimer();
      this.removeObservers();
    } catch(e) {};
    
    //destroy variables
    this.initialized = null;
  },

  getFileURI: function (aFile)
  {
    var ios = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
    // this.dump("getting URI for file " + aFile);
    var uri = ios.newFileURI(aFile);
    return uri.spec;
  },
  
  getWindow: function ffManager_getWindow()
  {
    //XXX may need to change this code if main window of a supported app is not "navigator:browser"
    var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService(Components.interfaces.nsIWindowMediator);
    var ww = Components.classes["@mozilla.org/embedcomp/window-watcher;1"].getService(Components.interfaces.nsIWindowWatcher);
    var top = wm.getMostRecentWindow("navigator:browser");
    if (!top) {
      switch (this.guid) {
        case "{ec8030f7-c20a-464f-9b0e-13a3a9e97384}":
        case "{3db10fab-e461-4c80-8b97-957ad5f8ea47}":
        default:        
          top = ww.openWindow(null, "chrome://browser/content/browser.xul", "_blank", "chrome,all,dialog=no", "about:blank");
          break;
        case "{86c18b42-e466-45a9-ae7a-9b95ba6f5640}":
          top = ww.openWindow(null, "chrome://navigator/content/navigator.xul", "_blank", "chrome,all,dialog=no", "about:blank");
          break;
      };
    };
    return top;  
  },

  removeFront: function(aArray, aNumEntries) {
    // remove aNumEntries from the front of aArray
    var newArray = [];
    for (var i=aNumEntries; i<aArray.length; i++) {
      newArray.push(aArray[i]);
    }
    return newArray;
  },
	
  receiveKnowledgeResponse: function(aXMLReq, success) {
    // success == 0 means failure to connect to agent, 1 is successful
    // response from agent.
    this.dump("receiveKnowledgeResponse(): success: " + success);
    if (success == 0) {
      this.dump("receiveKnowledgeResponse: Failed to connect to the " +
		"diagnostic agent.");
      return;
    }

    this.dump("Response: " + aXMLReq.responseText);

    // need to divide and multiply by 1000 to make it fit within a 32-bit int
    var knowledgeExpires = 1000 * this.prefs.getIntPref("extensions.diagtool.knowledgeExpires");
    var curTime = this.getCurrentTime();
    knowledgeExpires = curTime + KNOWLEDGE_LIFETIME;
    this.prefs.setIntPref("extensions.diagtool.knowledgeExpires",
			  knowledgeExpires / 1000);
    
    var doc = aXMLReq.responseXML;
    this.parseKnowledge(doc);
    this.writeKnowledgeBase();
  },

  receiveDiagResponse: function(aXMLReq, dest, success) {
    
    this.dump("Receiving diagnostic response. success: " + success);
    if (success == 0) {
      this.dump("Failed to connect to the diagnostic agent.");
      this.setToolbarDiagState("Failed to contact remote agent", true);
      return;
    }

    this.dump("Response: " + aXMLReq.responseText);
    var doc = aXMLReq.responseXML;
    var rootNode = doc.documentElement;
    var diagRespNodes = rootNode.getElementsByTagName('diagResponse');

    if (diagRespNodes.length != 1) {
      this.dump("receiveDiagResponse: Did not find diagResponse, diagRespNodes.length == " + diagRespNodes.length);
      this.setToolbarDiagState("Invalid response from diagnostic agent", true);
      return;
    }

    var diagRespNode = diagRespNodes[0];
    var responseStatus = diagRespNode.getAttribute("responseStatus");

    if (responseStatus == null) {
      this.dump("Error, did not get responseStatus from agent");
      this.setToolbarDiagState("No response from diagnostic agent", true);
      return;
    }

    if (responseStatus != 1) {
      this.dump("Error, diagnostic agent returned status " + status);
      this.setToolbarDiagState("Diagnosis failed", true);
      return;
    }

    this.parseObservations(doc);
    this.parseKnowledge(doc);

    var explanations = this.parseDiagExplanations(doc);
    if (explanations != null && explanations.length != 0) {
      this.displayDiag(dest, explanations[0], 'Regional agent');
      // diagnosis is complete, clear the state
      this.setToolbarDiagState('', true);
    } else {
      this.dump("No diag found for " + dest);
    }
  },

  receiveDiagHistNotificationResponse: function(aXMLReq, success) {
    // Handle a diag hist notification response from a regional agent

    // check the type of response, and then figure out what to do with
    // the data.
    this.dump("Receiving diag hist notification response. success: " +
	      success);
    if (success == 0) {
      this.dump("Failed to connect to the regional agent.");
      return;
    }

    this.dump("Response: " + aXMLReq.responseText);
    var doc = aXMLReq.responseXML;
    var rootNode = doc.documentElement;
    var responseNodes = rootNode.getElementsByTagName('notificationResponse');
    for (var respIdx=0; respIdx<responseNodes.length; respIdx++) {
      var respStatus = responseNodes[respIdx].getAttribute('responseStatus');
      if (respStatus != 1) {
	this.dump("receiveDiagHistNotificationResponse: Warning, Response status == " +
		  respStatus);
	return;
      }
    }

    // diagnosis completed successfully, clear the old history.
    //
    this.diagHist = this.removeFront(this.diagHist,
				     this.diagHistEntriesSent);
    this.diagHistEntriesSent = 0;
    
    this.writeDiagHist();
    
    this.parseObservations(doc);
    this.parseKnowledge(doc);
  },

  receiveHistoryNotificationResponse: function(aXMLReq, success) {
    // Receive a response from a diagnostic agent

    // check the type of response, and then figure out what to do with
    // the data.
    this.dump("Receiving connhist notification response. success: " + success);
    if (success == 0) {
      this.dump("Failed to connect to the history agent.");
      return;
    }

    this.dump("Response: " + aXMLReq.responseText);
    var doc = aXMLReq.responseXML;
    var rootNode = doc.documentElement;
    var responseNodes = rootNode.getElementsByTagName('notificationResponse');
    if (responseNodes.length == 0) {
      this.dump("receiveHistoryNotificationResponse: no responses found");
      return;
    }

    for (var respIdx=0; respIdx<responseNodes.length; respIdx++) {
      var respStatus = responseNodes[respIdx].getAttribute('responseStatus');
      if (respStatus != 1) {
	this.dump("receiveHistoryNotificationResponse: Warning, Response status == " +
		  respStatus);
	return;
      }
    }

    // diagnosis completed successfully, clear the old history.
    //
    var queue = this.connHist.entries;
    var entriesSent = this.connHist.entriesSent;
    this.connHist.entries = this.removeFront(queue, entriesSent);
    this.connHist.entriesSent = 0;
    this.writeConnHist();
    
    this.parseObservations(doc);
    this.parseKnowledge(doc);
  },

  receiveStatsResponse: function(aXMLReq, dest, success) {
    // Receive a response from a diagnostic agent

    // check the type of response, and then figure out what to do with
    // the data.
    this.dump("Receiving stats response. success: " + success);
    if (success == 0) {
      this.dump("Failed to connect to the stats agent.");
      this.alert("Stats temporarily unavailable, please try again later.");
      return;
    }

    this.dump("Response: " + aXMLReq.responseText);
    var doc = aXMLReq.responseXML;
    var elems = doc.getElementsByTagName("observationResponse");

    if (elems == null) {
      this.dump("Did not understand response from stats agent.");
      return;
    }

    if (elems.length == 1) {
      var status = elems[0].getAttribute('responseStatus')
      if (status != 1) {
	this.dump("Error, diagnostic agent returned status " + status);
	return;
      }
    } else {
      this.dump("Error, elems.length == " + elems.length);
      return;
    }

    var obsList = this.parseObservations(doc);
    this.parseKnowledge(doc);

    var data = this.parseDestInfo(obsList, dest);
    if (data != null) {
      this.displayStats(dest, data);
    } else {
      this.alert("No stats available for " + dest + " yet, try again later.");
    }
  },

  userFriendlyCause: function(cause) {
    // given a one letter cause such as 'l', return a user friendly
    // cause string such as 'local network failure'
    var causeMap = {
      'l': 'Local network failure',
      'w': 'Web server failure',
      'd': 'DNS lookup failure',
      'i': 'Other network failure'
    };
    if (cause in causeMap) {
      return causeMap[cause];
    } else {
      return 'Unknown';
    }
  },

  displayDiag: function(dest, explanation, source) {
    // display a diagnosis to the user.  Dest is the destination host
    // to which the user couldn't connect.  explanation is a
    // cause/prob associative array.

    var rootCause = this.userFriendlyCause(explanation.cause);
    var prob = (Math.round(explanation.prob * 10000) / 100) + '%';
    
    this.setToolbarDiagResult(rootCause + '|' +
			     'Confidence: ' + prob + '\n' + 
			     'Diagnosis by ' + source + '\n'
			     );
  },

  displayStats: function(dest, data) {
    var win = this.getWindow();
    var response = "chrome://diag/content/response.xul";
    win.openDialog(response,
		   "Server Info - " + dest,
		   "chrome,all,dialog=no",
		   data);
  },

  parseDiagExplanations: function(doc)
  {
    // parse a diagnostic response to extract a probabilistic
    // diagnosis.

    // note that the componentMap here depends on id of the components
    // in the diagRequest.  If you renumber the components you have to
    // update the map.
    var componentMap =
    {
      '2': 'l',
      '3': 'w',
      '4': 'd',
      '5': 'i',
    };
      
    var explanations = new Array();

    var explanationNodes = doc.getElementsByTagName('explanation');
    for (var explIndex=0; explIndex<explanationNodes.length; explIndex++) {
      // assume that explanations are in order of most likely to least likely
      var explanationNode = explanationNodes[explIndex];
      var likelihood = explanationNode.getAttribute('p');
      var componentRefs = explanationNode.getElementsByTagName('componentRef');
      var numFailedComponents = 0;
      for (var crIndex=0; crIndex<componentRefs.length; crIndex++) {
	var componentRefNode = componentRefs[crIndex];
	var crId = componentRefNode.getAttribute('ref');
	var crStatus = componentRefNode.getAttribute('status');
	if (crStatus == 'FAIL') {
	  // assume only one component has failed
	  numFailedComponents++;
	  var explanation =
	    {
	      cause: componentMap[crId],
	      prob: likelihood,
	    };
	  explanations.push(explanation);
	}  
      }
      if (numFailedComponents != 1) {
	this.dump("numFailedComponents != 1.  numFailedComponents == " +
		  numFailedComponents);
      }
    }
    return explanations;
  },

  parseDestInfo: function(obsList, dest)
  {
    // return an associative array of the information in destInfo
    //
    if (obsList.length != 1) {
      this.dump("parseDestInfo: got unexpected number of obs: " +
		obsList.length);
      return;
    }

    var obs = obsList[0];

    var data = {};

    var consecFailures = null;

    // this.dump('obs.props: ' + obs.props.toSource());

    var SS_PREFIX = 'http://capri.csail.mit.edu/2006/capri/servstats#';
    data.users = parseInt(obs.props[SS_PREFIX + "users"]);

    if ((SS_PREFIX + "rootCause") in obs.props) {
      data.rootCause = obs.props[SS_PREFIX + "rootCause"];
    } else {
      data.rootCause = "---";
    }

    if ((SS_PREFIX + "consecFailures") in obs.props) {
      data.consecFailures = parseInt(obs.props[SS_PREFIX + 'consecFailures']);
    } else {
      data.consecFailures = "---";
    }

    if ((SS_PREFIX + "avgLatency") in obs.props) {
      data.avgLatency = parseInt(obs.props[SS_PREFIX + "avgLatency"]);
      data.avgLatency = (data.avgLatency / 1000).toFixed(3) + " s";
    } else {
      data.avgLatency = "---";
    }

    if ((SS_PREFIX + "lastSuccess") in obs.props) {
      var lastSuccessTime = parseInt(obs.props[SS_PREFIX + "lastSuccess"]);
      data.lastSuccess = this.dateToString(lastSuccessTime);
    } else {
      data.lastSuccess = "---";
    }

    if ((SS_PREFIX + "lastFailure") in obs.props) {
      var lastFailureTime = parseInt(obs.props[SS_PREFIX + "lastFailure"]);
      data.lastFailure = this.dateToString(lastFailureTime);
    } else {
      data.lastFailure = "---";
    }
    
    var statusDist = {};
    var failures = 0;
    var conns = 0;
    var recentStatusDist = obs.props[SS_PREFIX + "recentStatusDist"];
    recentStatusDist = this.trim(recentStatusDist);
    if (recentStatusDist != '') {
      var statuses = recentStatusDist.split('|');
      for (var i=0; i<statuses.length; i++) {
	var statusParts = statuses[i].split(',');
	var status = parseInt(statusParts[0]);
	var count  = parseInt(statusParts[1]);
	conns += count;
	if (status != OK) {
	  failures += count;
	}
	if (!(status in statusDist)) {
	  statusDist[status] = 0;
	}
	statusDist[status] += count;
      }
    }

    data.host = dest;
    if (conns > 0) {
      data.failureRate = (failures / conns * 100).toFixed(2) + "%";
    } else {
      data.failureRate = "---";
    }
    data.conns = conns;
    data.statusDist = statusDist;
    
    return data;
  },

  createObservation: function(doc,
			      rootNode,
			      obsType,
			      componentId,
			      className,
			      props,
			      metadata) {
    // Append an observation to the rootNode of document doc
    //
    var observationNode = doc.createElement('observation');
    observationNode.setAttribute('time', metadata['time']);
    observationNode.setAttribute('originator', metadata['originator']);
    var componentNode = doc.createElement(obsType);
    componentNode.setAttribute('id', componentId);
	
    var classNode = doc.createElement('class');
    var classText = doc.createTextNode(className);
    classNode.appendChild(classText);
    componentNode.appendChild(classNode);
	
    for (var propId=0; propId<props.length; propId++) {
      var childNode = null;
      var prop = props[propId];
      
      if (prop.propVal != undefined &&
	  prop.propVal != null &&
	  prop.propVal != '') {
	childNode = doc.createTextNode(prop.propVal);
      } else if (prop.propRef != undefined && prop.propRef != null) {
	childNode = doc.createElement('componentRef');
	childNode.setAttribute('ref', prop.propRef);
      }

      if (childNode != null) {
	var propNode = doc.createElement(prop.propName);
	propNode.appendChild(childNode);
	componentNode.appendChild(propNode);
      }
    }
    observationNode.appendChild(componentNode);
    rootNode.appendChild(observationNode);
  },

  diagHistToNotifications: function() {
    // generate an XML list of observations from diagHist
    var notifications = [];
    var diagHist = this.diagHist;
    var componentId = 1;
    var totalCffu = this.totalCffu;
    for (var i=0; i<diagHist.length; i++) {
      var diagEntry = diagHist[i];
      var text = '<body>';
      text += this.serializeFailureObservation(componentId,
					       diagEntry);
      text += '</body>';
      var notification = this.constructMessage('notification',
					       'notify:diagHist', text);
      notifications.push(notification);
      componentId += 10; // minimum 7, but I use 10 just to be clearer
    }
    return notifications;
  },

  requestAgentMap: function(reload) {
    // Get a desthash to history agent mapping from a directory
    // server.  Mapping is by first byte of the desthash.
    var curTime = this.getCurrentTime();
    var retryTime = 1000 * this.prefs.getIntPref("extensions.diagtool.retryTime");
    
    if (!reload && curTime < retryTime) {
      // wait until a retry time before requesting again
      return;
    }

    var text = '';
    text += '<agentMapRequest';
    text += ' agentType="user"\n';
    text += ' version="' + PROTO_VER + '"\n';
    text += ' time="' + curTime + '"\n';
    text += ' clientAgent="' + CLIENT_AGENT + '"\n';
    text += ' clientVer="' + CLIENT_VER + '"\n';
    text += '/>\n';

    var agent = this;
    var onLoad = function() {
      agent.receiveAgentMapResponse(this, 1);
    };
    var onError = function() {
      agent.receiveAgentMapResponse(this, 0);
    };

    var directoryServerURI = this.prefs.getCharPref("extensions.diagtool.directoryServerURI");

    this.sendMessage(directoryServerURI, text, onLoad, onError);

    // send no more than one request every retry_timeout
    var nextRetryTime = curTime + RETRY_TIMEOUT;
    this.prefs.setIntPref("extensions.diagtool.retryTime",
			  nextRetryTime / 1000);
  },
    
  receiveAgentMapResponse: function(aXMLReq, status) {
    if (status == 0) {
      // error.
      this.dump("Error getting agent map.");
      return;
    }

    // need to divide and multiply by 1000 to make it fit within a 32-bit int
    var agentMapExpires = 1000 * this.prefs.getIntPref("extensions.diagtool.agentMapExpires");
    var curTime = this.getCurrentTime();
    agentMapExpires = curTime + AGENT_MAP_LIFETIME;
    this.prefs.setIntPref("extensions.diagtool.agentMapExpires",
			  agentMapExpires / 1000);
    var doc = aXMLReq.responseXML;
    this.dump('Received agent map:\n' + this.serializeXMLDoc(doc));
    this.agentMapDoc = doc;
    this.parseAgentMap(doc);
  },

  parseAgentMap: function(doc) {
    // parse a hist agent map DOM document and set the
    // regionalAgentURI and load the historyAgentMap.

    var adNodes = doc.getElementsByTagName('serviceAdvertisement');
    var diagServices = [];
    for (var adNodesIdx=0; adNodesIdx<adNodes.length; adNodesIdx++) {
      var adNode = adNodes[adNodesIdx];
      var attrList = ['messageType', 'agentURI', 'serviceID', 'cost'];
      var svc = {};
      for (var i=0; i<attrList.length; i++) {
	var attr = attrList[i];
	svc[attr] = adNode.getAttribute(attr);
      }
      
      this.dump('parseAgentMap: (' + svc.agentURI + ', ' + svc.messageType +
		', ' + svc.serviceID + ', ' + svc.cost + ')');

      // Adjust cost to ensure that the sort below is stable.
      // Assumes that the difference between any two costs is less
      // than the number of services / 1000
      svc.cost = svc.cost + (adNodesIdx / 1000.0);
      
      // Currently all observation notifications and knowledge
      // requests also go to the regional agent.  If this changes in
      // the future, then we will need to parse the other types of
      // messages as well.
      if (svc.messageType == 'diagRequest') {
	diagServices.push(svc);
      }
    }
    
    function costComparison(a, b) {
      // sort in ascending order of cost.
      if (a.cost < b.cost) {
	return -1;
      } else if (a.cost == b.cost) {
	return 0;
      } else if (a.cost > b.cost) {
	return 1;
      }
    }
    // Javascript sort is not stable, but we ensured that the costs
    // are all different when we set svc.cost above.
    diagServices.sort(costComparison);
    
    if (diagServices.length > 0) {
      // Use the lowest cost service in the list.  In case of a tie,
      // use the first one in the list.
      var svc = diagServices[0];
      this.dump('parseAgentMap: Using regional agent ' + svc.agentURI);
      this.prefs.setCharPref("extensions.diagtool.regionalAgentURI",
			     svc.agentURI);
      this.prefs.setCharPref("extensions.diagtool.diagServiceID",
			     svc.serviceID);
      
      // if we don't have a knowledge base yet, request some
      // knowledge now that we have a new agent.
      if (this.knowledgeBase.length == 0) {
	this.sendKnowledgeRequest();
      }
    } else {
      this.dump('parseAgentMap: No diag services found');
    }

    this.writeAgentMap();
  },

  connHistToNotifications: function() {
    // generate an XML list of observations from connHist.  Note that
    // the history agent to whom to send a message depends on the
    // hosthash.
    //
    // Return an associative array of agentURI/xml messages.
    //
    var connHist = this.connHist;

    var componentId = 1;
    var notifications = [];

    var queue = connHist.entries;
    for (var i=0; i<queue.length; i++) {
      var doc = this.createXMLDoc('body');
      var rootNode = doc.documentElement;
      var entry = queue[i];

      var httpComponentId = componentId++;
      var ffoxComponentId = componentId++;
	
      var httpProps = [
	{
	  propName: 'com:destHash',
	  propVal: entry.hostHash
	},
	{
	  propName: 'com:destIP',
	  propVal: entry.destIP
	},
	{
	  propName: 'com:srcIP',
	  propVal: SRC_IP
	},
	{
	  propName: 'com:connTime',
	  propVal: entry.connTime
	},
	{
	  propName: 'core:status',
	  propVal: entry.status
	},
	{
	  propName: 'com:elapsedTime',
	  propVal: entry.elapsedTime
	},
	{
	  propName: 'com:ffoxErrorTest',
	  propRef: ffoxComponentId
	}
	];

      // add a firefox error observation
      var ffoxProps = [
	{
	  propName: 'com:ffoxErrorCode',
	  propVal: entry.errorCode,
	},
	{
	  propName: 'com:httpConn',
	  propRef: httpComponentId,
	}
	];
      
      var metadata = {
	'time': entry.connTime,
	'originator': 'userAgent'
      };

      this.createObservation(doc, rootNode, 'component', httpComponentId,
			     PREFIX + 'common#HTTP_Connection',
			     httpProps, metadata);
	
      this.createObservation(doc, rootNode, 'test', ffoxComponentId,
			     PREFIX + 'common#Firefox_Error_Test',
			     ffoxProps, metadata);
      var xml = this.serializeXMLDoc(doc);
      var notification = this.constructMessage('notification',
					       'notify:connHist', xml);
      notifications.push(notification);
    }

    return notifications;
  },

  serializeConnHist: function() {
    var connHist = this.connHist;
    var doc = this.createXMLDoc('connHist');
    var rootNode = doc.documentElement;

    var connHistQueue = connHist.entries;
    var connHistHash = {};
    // construct a hash of queues
    for (var entryIndex=0; entryIndex<connHistQueue.length; entryIndex++) {
      var entry = connHistQueue[entryIndex];
      if (!(entry.hostHash in connHistHash)) {
	connHistHash[entry.hostHash] = [];
      }
      connHistHash[entry.hostHash].push(entry);
    }
	
    for (var hostHash in connHistHash) {
      var hostNode = doc.createElement("host");
      hostNode.setAttribute("hash", hostHash);
      rootNode.appendChild(hostNode);
    // this.dump("Hash: " + hostHash + ", conns: " + connHist[hostHash].length);
      for (var i=0; i<connHistHash[hostHash].length; i++) {
        var entry = connHistHash[hostHash][i];
        var connNode = doc.createElement("conn");
        if ("elapsedTime" in entry) {
          connNode.setAttribute("et", entry.elapsedTime);
        }
        connNode.setAttribute("st", entry.status);
        connNode.setAttribute("ec", entry.errorCode);
        connNode.setAttribute("ct", entry.connTime);
	if (entry.destIP != undefined) {
	  connNode.setAttribute("ip", entry.destIP);
	}
        hostNode.appendChild(connNode);
      }
    }

    var xml = this.serializeXMLDoc(doc);
    return xml;
  },

  createXMLDoc: function(rootElement) {
    var parser = Components.classes["@mozilla.org/xmlextras/domparser;1"].createInstance(Components.interfaces.nsIDOMParser);           
    var doc = parser.parseFromString("<" + rootElement + "/>", "text/xml");
    return doc;
  },
  
  serializeXMLDoc: function(doc) {
    var serializer = Components.classes["@mozilla.org/xmlextras/xmlserializer;1"].createInstance(Components.interfaces.nsIDOMSerializer);

    var xml = serializer.serializeToString(doc);
    return xml;
  },


  serializeKnowledgeBase: function() {
    // serialize dependency knowledge as XML.  This assumes that each
    // piece of knowledge in the knowledge base already has a DOM node
    // associated with it, which will be the case for all knowledge
    // read in externally from files or from other agents.
    var rootElement = 'knowledgeBase';
    var doc = this.createXMLDoc(rootElement);

    var knowledgeBase = this.knowledgeBase;
    var rootNode = doc.getElementsByTagName(rootElement)[0];
    for (var propertyName in knowledgeBase) {
      var propertyKnowledge = knowledgeBase[propertyName];
      var knowledgeNode = doc.importNode(propertyKnowledge.node, true);
      rootNode.appendChild(knowledgeNode);
    }
    var xml = this.serializeXMLDoc(doc);
    return xml;
    
  },

  serializeDiagHist: function() {
    var diagHist = this.diagHist;
    var rootElement = "diagHist";
    var doc = this.createXMLDoc(rootElement);

    var rootNode = doc.documentElement;
    for (var i=0; i<diagHist.length; i++) {
      var diagNode = doc.createElement("diag");
      var entry = diagHist[i];
      var diagAttributes = this.diagAttributes;

      for (var diagAttrId=0; diagAttrId<diagAttributes.length; diagAttrId++) {
	var attribute = diagAttributes[diagAttrId];
	if (entry[attribute] != null) {
	  diagNode.setAttribute(attribute,          entry[attribute].value);
	  diagNode.setAttribute(attribute + 'Time', entry[attribute].time);
	  rootNode.appendChild(diagNode);
	}
      }
      
//       this.dump("destHash: "  + entry.destHash +
// 		"diagTime: "  + entry.diagTime +
// 		"connTime: "  + entry.connTime +
// 		"rootCause: " + entry.rootcause +
// 		"usingEvidence: " + entry.usingEvidence);
    }

    var xml = this.serializeXMLDoc(doc);
    return xml;
  },

  md5sum: function(s) {
    // for additional privacy, the user can set the salt.  If you only
    // wish to share diagnoses among your friends, for example, you
    // can all use the same secret salt.
    // base64 determines whether the result is base64 encoded.
    var privGroup = this.prefs.getIntPref("extensions.diagtool.privacyGroup");
    var salt = "";
    if (privGroup == 1) {
      salt = this.prefs.getCharPref("extensions.diagtool.hashSalt");
    }
    s = salt + DATA_ID + s;

    var stream = Components.classes["@mozilla.org/io/string-input-stream;1"].createInstance(Components.interfaces.nsIStringInputStream);
    stream.setData(s, s.length);
    const nsICryptoHash = Components.interfaces.nsICryptoHash;
    var hashComp = Components.classes["@mozilla.org/security/hash;1"].createInstance(nsICryptoHash);
    
    hashComp.init(nsICryptoHash.MD5);
    hashComp.updateFromStream(stream, stream.available());
    var base64 = true;
    var hash = hashComp.finish(base64); // encode as base64
    return hash;
    return hashComp;
  },

  constructCapriMessage: function(messages) {
    // given an array of messages, combine them all together into one
    // big CAPRI message.
    var text = '';
    var t = this.getCurrentTime();
    text += '<?xml version="1.0" encoding="UTF-8" ?>\n';
    text += '<capriMessage\n';
    text += ' xmlns:core="http://capri.csail.mit.edu/2006/capri/core#"\n';
    text += ' xmlns:com="http://capri.csail.mit.edu/2006/capri/common#"\n';
    text += ' version="' + PROTO_VER + '"\n';
    text += ' time="' + t + '"';
    text += ' clientAgent="' + CLIENT_AGENT + '"\n';
    text += ' clientVer="' + CLIENT_VER + '"\n';
    text += '>\n';
    text += messages.join('\n');
    text += "</capriMessage>\n";
    return text;
  },    

  constructMessage: function(messageType, serviceId, contents) {
    var t = this.getCurrentTime();
    var text = '';
    text += '<' + messageType;
    var idAttrName;
    var idVal;
    if (messageType == 'notification') {
      idAttrName = 'notificationID';
      idVal = this.reqId++;
    } else if (messageType == 'observationRequest' ||
	       messageType == 'diagRequest' ||
	       messageType == 'knowledgeRequest') {
      idAttrName = 'requestID';
      idVal = this.reqId++;
    } else {
      this.dump('Unknown messageType: ' + messageType);
    }
    
    text += ' serviceID="' + serviceId + '"';
    text += ' ' + idAttrName + '="' + idVal + '"';
    text += '>\n';
    text += contents;
    text += "</" + messageType + ">\n";
    return text;
  },
    
  createStatsRequest: function(destHash) {
    // given a desthash, create a stats request
    //
    var text = '';
    text += '<inputComponent>\n';
    text += '  <component>\n';
    text += '    <class>\n';
    text += '      http://capri.csail.mit.edu/2006/capri/common#HTTP_Connection\n';
    text += '    </class>\n';
    text += '    <com:destHash>' + destHash + '</com:destHash>\n';
    text += '  </component>\n';
    text += '</inputComponent>\n';
    text += '<body />\n';

    var message = this.constructMessage('observationRequest', 'obs:stats', text);
    return [message];
  },

  serializeFailureObservation: function(componentId, failureObservations) {
    // Creates the xml for a set of observations corresponding to a
    // failure.
    //
    // componentId is the number of the next unused componentId.  This
    // procedure requires that all IDs within the interval
    // [componentId, componentId + 6] not be reused.  In other words,
    // for successive calls to this function, componentId should
    // differ by at least 7.
    //
    // failureObservations is an associative array with keys destHash,
    // cffu, cfts, ffoxErrorCode, probeURI, probeResult, and diagResult.  
    // Each failureObservation value is an associative array of
    // time/value entries.
    //
    // all keys must be defined with the exception of cfts, probeURI,
    // probeResult, dnsHostname, dnsIPAddrs, and diagResult, which are
    // optional.
    //
    // Returns the xml text of the corresponding observations.
    //
    
    // httpConnection ID must be the lowest numbered one for the
    // inputComponent reference to be correct.  These numbers must
    // match with the explanation numbers in createDiagRequest().
    var httpConnectionId = componentId++;
    var localNetId = componentId++;
    var webServerId = componentId++;
    var dnsLookupId = componentId++;
    var ipRoutingId = componentId++;
    var obconnId = componentId++;
    var ffoxId = componentId++;
    var diagResultId = componentId++;
    if (failureObservations.destHash == null) {
      // No info about this failure available... (well, perhaps we
      // have some info, but...
      this.dump("Error, did not find destHash in failureObservations (perhaps this was not a real diagnosis?)");
      return '';
    }

    var text = '';
    text += '<observation time="' + failureObservations.destHash.time + '">\n';
    text += '  <component id="' + httpConnectionId + '">\n';
    text += '    <class>\n';
    text += '      http://capri.csail.mit.edu/2006/capri/common#HTTP_Connection\n';
    text += '    </class>\n';
    text += '    <com:srcIP>' + SRC_IP + '</com:srcIP>\n';
    if (this.sendHostname &&
	failureObservations.destIP != null &&
	failureObservations.destIP.value != null) {
      text += '    <com:destIP>' + failureObservations.destIP.value + '</com:destIP>\n';
    }
    text += '    <com:destHash>' + failureObservations.destHash.value + '</com:destHash>\n';
    text += '    <com:connTime>' + failureObservations.destHash.time + '</com:connTime>\n';
    if (failureObservations.elapsedTime != null &&
	failureObservations.elapsedTime.value != null) {
      text += '    <com:elapsedTime>' + failureObservations.elapsedTime.value + '</com:elapsedTime>\n';
    }
    text += '    <core:status>FAIL</core:status>\n';
    text += '    <com:dnsLookup><componentRef ref="' + dnsLookupId + '"/></com:dnsLookup>\n';
    text += '    <com:ipRouting><componentRef ref="' + ipRoutingId + '"/></com:ipRouting>\n';
    text += '    <com:httpServer><componentRef ref="' + webServerId + '"/></com:httpServer>\n';
    text += '    <com:localNet><componentRef ref="' + localNetId + '"/></com:localNet>\n';
    text += '    <com:ffoxErrorTest><componentRef ref="' + ffoxId + '"/></com:ffoxErrorTest>\n';
    text += '    <com:outboundConnTest><componentRef ref="' + obconnId + '"/></com:outboundConnTest>\n';
    text += '  </component>\n';
    text += '</observation>\n';
    
    text += '<observation time="' + failureObservations.cffu.time + '">\n';
    text += '  <component id="' + localNetId + '">\n';
    text += '    <class>\n';
    text += '      http://capri.csail.mit.edu/2006/capri/common#Local_Network\n';
    text += '    </class>\n';
    text += '    <com:ipAddr>' + SRC_IP + '</com:ipAddr>\n';
    text += '    <com:consecFailuresFromUser>' + failureObservations.cffu.value + '</com:consecFailuresFromUser>\n';
    if (failureObservations.totalCffu != null) {
      text += '    <com:totalConsecFailuresFromUser>' + failureObservations.totalCffu.value + '</com:totalConsecFailuresFromUser>\n';
    }
    text += '  </component>\n';
    text += '</observation>\n';
    text += '\n';

    var webServerTime;
    if (failureObservations.cfts != null &&
	failureObservations.cfts.time != null) {
      webServerTime = failureObservations.cfts.time;
    } else {
      webServerTime = failureObservations.destHash.time;
    }
    
    text += '<observation time="' + webServerTime + '">\n';
    text += '  <component id="' + webServerId + '">\n';
    text += '    <class>\n';
    text += '      http://capri.csail.mit.edu/2006/capri/common#HTTP_Server\n';
    text += '    </class>\n';
    text += '    <com:hostHash>' + failureObservations.destHash.value + '</com:hostHash>\n';
    text += '    <com:connTime>' + failureObservations.destHash.time + '</com:connTime>\n';
    text += '    <com:httpConn><componentRef ref="' + httpConnectionId + '"/></com:httpConn>\n';
    if (this.sendHostname &&
	failureObservations.destIP != null &&
	failureObservations.destIP.value != null) {
      text += '    <com:ipAddr>' + failureObservations.destIP.value + '</com:ipAddr>\n';
    }

    if (failureObservations.cfts != null &&
	failureObservations.cfts.value != null) {
      text += '    <com:consecFailuresToServer>' + failureObservations.cfts.value + '</com:consecFailuresToServer>\n';
    }

    text += '  </component>\n';
    text += '</observation>\n';
    text += '\n';
    text += '<observation time="' + failureObservations.destHash.time + '">\n';
    text += '  <component id="' + dnsLookupId + '">\n';
    text += '    <class>\n';
    text += '      http://capri.csail.mit.edu/2006/capri/common#DNS_Lookup\n';
    text += '    </class>\n';
    text += '    <com:srcIP>' + SRC_IP + '</com:srcIP>\n';
    text += '    <com:destHash>' + failureObservations.destHash.value + '</com:destHash>\n';

    if (this.sendHostname &&
	failureObservations.dnsHostname != null &&
	failureObservations.dnsHostname.value != null) {
      text += '    <com:hostname>' + failureObservations.dnsHostname.value + '</com:hostname>\n';
      if (failureObservations.dnsIPAddrs != null &&
	  failureObservations.dnsIPAddrs.value != null) {
	text += '    <com:ipAddrs>' + failureObservations.dnsIPAddrs.value + '</com:ipAddrs>\n';
      }
    }
	

    text += '  </component>\n';
    text += '</observation>\n';
    text += '\n';
    text += '<observation time="' + failureObservations.destHash.time + '">\n';
    text += '  <component id="' + ipRoutingId + '">\n';
    text += '    <class>\n';
    text += '      http://capri.csail.mit.edu/2006/capri/common#IP_Routing\n';
    text += '    </class>\n';
    text += '    <com:srcIP>' + SRC_IP + '</com:srcIP>\n';

    if (this.sendHostname &&
	failureObservations.destIP != null &&
	failureObservations.destIP.value != null) {
      text += '    <com:destIP>' + failureObservations.destIP.value + '</com:destIP>\n';
    }

    text += '    <com:destHash>' + failureObservations.destHash.value + '</com:destHash>\n';
    text += '  </component>\n';
    text += '</observation>\n';
    text += '\n';

    if (failureObservations.probeResult != null &&
	failureObservations.probeURI != null &&
	failureObservations.probeResult.time != null) {
      text += '<observation time="' + failureObservations.probeResult.time + '">\n';
      text += '  <test id="' + obconnId + '">\n';
      text += '    <class>\n';
      text += '      http://capri.csail.mit.edu/2006/capri/common#Outbound_Conn_Test\n';
      text += '    </class>\n';
      text += '    <com:srcIP>' + SRC_IP + '</com:srcIP>\n';
      text += '    <com:probeURI>' + failureObservations.probeURI.value + '</com:probeURI>\n';
      text += '    <com:probeResult>' + failureObservations.probeResult.value + '</com:probeResult>\n';
      text += '    <com:localNet><componentRef ref="' + localNetId + '"/></com:localNet>\n';
      text += '  </test>\n';
      text += '</observation>\n';
      text += '\n';
    }

    text += '<observation time="' + failureObservations.destHash.time + '">\n';
    text += '  <test id="' + ffoxId + '">\n';
    text += '    <class>\n';
    text += '      http://capri.csail.mit.edu/2006/capri/common#Firefox_Error_Test\n';
    text += '    </class>\n';
    text += '    <com:srcIP>' + SRC_IP + '</com:srcIP>\n';
    text += '    <com:destHash>' + failureObservations.destHash.value + '</com:destHash>\n';
    text += '    <com:ffoxErrorCode>' + failureObservations.ffoxErrorCode.value + '</com:ffoxErrorCode>\n';
    text += '    <com:httpConn><componentRef ref="' + httpConnectionId + '"/></com:httpConn>\n';
    text += '  </test>\n';
    text += '</observation>\n';

    if (failureObservations.diagResult != null) {
      var failedComponentMap = {
	'l' : localNetId,
	'w' : webServerId,
	'd' : dnsLookupId,
	'i' : ipRoutingId,
      };

      var explanation = failureObservations.diagResult.value;
      this.dump('explanation == ' + explanation);
      var parts = explanation.split(',');
      var cause = parts[0];
      var failProb = parts[1];
      var okProb = 1.0 - failProb;
      if (cause in failedComponentMap) {
	var failedComponentId = failedComponentMap[cause];
      } else {
	this.dump("Error, could not find cause " + cause);
      }
    
      text += '<belief';
      text += ' time="' + failureObservations.diagResult.time + '"';
      text += ' originator="userAgent" >\n';
      text += '  <subject><componentRef ref="' + failedComponentId + '"/></subject>\n';
      text += '  <property>http://capri.csail.mit.edu/2006/capri/core#status</property>\n';
      text += '  <distribution>\n';

      text += '    <entry value="OK"   p="' + okProb   + '" />\n';
      text += '    <entry value="FAIL" p="' + failProb + '" />\n';
      text += '  </distribution>\n';
      text += '</belief>\n';
    }
    return text;
  },

  getFailureObservationsFromDepGraph: function(depGraph) {
    // given a dependency graph, return an associative array of
    // failure observations.  This is the only info about a failure
    // that gets sent to other agents.  This also gets logged into
    // diagHist.
    var failureObservations = {
      destHash:      depGraph.getProperty('common#Firefox_Error_Test',
					  'common#destHash'),
      elapsedTime:   depGraph.getProperty('common#HTTP_Connection',
					  'common#elapsedTime'),
      cffu:          depGraph.getProperty('common#Local_Network',
					  'common#consecFailuresFromUser'),
      cfts:          depGraph.getProperty('common#HTTP_Server',
					  'common#consecFailuresToServer'),
      probeResult:   depGraph.getProperty('common#Outbound_Conn_Test',
					  'common#probeResult'),
      probeURI:      depGraph.getProperty('common#Outbound_Conn_Test',
					  'common#probeURI'),
      dnsHostname:   depGraph.getProperty('common#DNS_Lookup',
					  'common#hostname'),
      dnsIPAddrs:    depGraph.getProperty('common#DNS_Lookup',
					  'common#ipAddrs'),
      destIP:        depGraph.getProperty('common#IP_Routing',
					  'common#destIP'),
      ffoxErrorCode: depGraph.getProperty('common#Firefox_Error_Test',
					  'common#ffoxErrorCode'),
    };
    return failureObservations;
  },
  
  createDiagRequest: function(destHash, depGraph) {
    // create a diagnostic request.  destHash should match the
    // depGraph ffox_error_test desthash value if one exists.
    //
    if (depGraph == null) {
      this.dump("depGraph is null!");
      return;
    }

    var failureObservations = this.getFailureObservationsFromDepGraph(depGraph);

    // Note that these componentRef numbers below must match the
    // component numbers in serializeFailureObservation().
    var componentId = 1;
    var text = ''
    text += '<inputComponent>\n'
    text += '<componentRef ref="1" />\n';
    text += '</inputComponent>\n';
    text += '<explanations>\n';
    text += '  <explanation>\n';
    text += '    <componentRef ref="2" status="FAIL" />\n';
    text += '  </explanation>\n';
    text += '  <explanation>\n';
    text += '    <componentRef ref="3" status="FAIL" />\n';
    text += '  </explanation>\n';
    text += '  <explanation>\n';
    text += '    <componentRef ref="4" status="FAIL" />\n';
    text += '  </explanation>\n';
    text += '  <explanation>\n';
    text += '    <componentRef ref="5" status="FAIL" />\n';
    text += '  </explanation>\n';
    text += '</explanations>\n';

    text += '<body>\n';
    text += this.serializeFailureObservation(componentId, failureObservations);
    text += '</body>\n';

    var message = this.constructMessage('diagRequest', 'diag:http', text);
    return [message];
  },

  setNodeAttributes: function(node, attributeNames) {
    // Given a node in a dependency graph, set the values of the given
    // attributes of that component: range, parents, children, and cpt.
    //
    var nodeAttributes = this.nodeAttributes;

    var componentName = node.name;
    for (var i=0; i<attributeNames.length; i++) {
      var attributeName = attributeNames[i];
      var value = nodeAttributes[componentName][attributeName];
      if (value != undefined) {
	node[attributeName] = value;
      }
    }
  },
    
  buildNode: function(bnet, name) {
    var node = new bnNode();
    node.agent = this; // for convenience.
    node.bnet = bnet;
    bnet[name] = node;
    node.name = name;
    this.setNodeAttributes(node, ['range', 'parents', 'children', 'cpt']);
    return node;
  },

  buildBNetFromDepGraph: function(depGraph) {
    // Return a dependency graph that models root cause of failure,
    // outbound connectivity, consecutive failures from user, http
    // conn status, consecutive failures to server, and firefox error
    // code.
    //
    // You should rebuild the dependency graph every time you get new
    // knowledge.
    //
    var bnet = {};
    var names = ['cause', 'obconn', 'http', 'cffu', 'http', 'cfts', 'ffox'];
    for (var i=0; i<names.length; i++) {
      this.buildNode(bnet, names[i]);
    }

    var evidence = this.extractEvidenceFromDepGraph(depGraph);
    this.addEvidenceToBNet(bnet, evidence);

    return bnet;
  },

  addEvidenceToBNet: function(bnet, evidence) {
    // given a dependency graph/bayes net, add evidence to it.
    // evidence is an associative array with entries of the form
    // evidence[nodeName] == nodeValue.
    var allEvidence = ''
    for (var nodeName in evidence) {
      var value = evidence[nodeName];
      if (bnet[nodeName].range.indexOf(value) == -1) {
	this.dump('Warning, value ' + value + ' not in range for ' + nodeName);
      }
      allEvidence += nodeName + ': ' + value + '\n';
      bnet[nodeName].setEvidence(value);
    }
    this.dump(allEvidence);
  },

  updateBeliefs: function(bnet) {
    var cause = bnet['cause'];
    for (var childId=0; childId<cause.children.length; childId++) {
      var childName = cause.children[childId];
      bnet[childName].calcLambda();
      bnet[childName].calcLambdaMessages();
    }

    cause.updateBelief();
  
    for (var childId=0; childId<cause.children.length; childId++) {
      var childName = cause.children[childId];
      bnet[childName].updateBelief();
    }
  },    

  logBeliefs: function(bnet) {
    var rv = '';
    for (var nodeName in bnet) {
      var node = bnet[nodeName];
      for (var valId=0; valId<node.range.length; valId++) {
	var val = node.range[valId];
	rv += nodeName + '.bel[' + val + '] == ' + Math.round(node.bel[val] * 10000) / 10000 + '\n';
      }
    }
    this.dump(rv);
  },

  verifyProbs: function(correctProbs, explanations) {
    // given an associative array correctProbs mapping cause to
    // correct explanations, check if the explanation probs match
    var testPassed = true;
    for (var i=0; i<explanations.length; i++) {
      var cause = explanations[i].cause;
      var explProb = Math.round(explanations[i].prob * 10000);
      if (explProb != correctProbs[cause]) {
	testPassed = false;
	this.dump("Test FAILED, Pr(" + cause + ") == " + explProb +
		  " != " + correctProbs[cause]);
      }
    }
    if (testPassed) {
      this.dump("Verify probs test passed");
    }
    return testPassed;
  },
  
  inferenceTest: function() {
    // run several types of inference tests.
    //

    var depGraph = this.initDependencyGraph();
    this.dump("with cffu:");
    depGraph.setProperty('common#Local_Network',
			 'common#consecFailuresFromUser',
			 curTime,
			 1);
    var explanations = this.inferRootCause(depGraph);
    var correctProbs = {l: 6522, w: 500, d: 1000, i: 1200};
    this.verifyProbs(correctProbs, explanations);

    this.dump("with cffu:");
    depGraph.setProperty('common#HTTP_Connection',
			 'core#status',
			 curTime,
			 'FAIL');
    var explanations = this.inferRootCause(depGraph);
    correctProbs = {l: 8834, w: 677, d: 1354, i: 1625};
    this.verifyProbs(correctProbs, explanations);

    depGraph = this.initDependencyGraph();
    this.dump("with cffu and ffox:");
    depGraph.setProperty('common#Local_Network',
			 'common#consecFailuresFromUser',
			 curTime,
			 1);
    depGraph.setProperty('common#Firefox_Error_Test',
			 'common#ffoxErrorCode',
			 curTime,
			 14);
    var explanations = this.inferRootCause(depGraph);
    correctProbs = {l: 9047, w: 694, d: 277, i: 1665};
    this.verifyProbs(correctProbs, explanations);


    ///////////////////////////////////////////////
    depGraph = this.initDependencyGraph();
    depGraph.setProperty('common#HTTP_Connection',
			 'core#status',
			 curTime,
			 'FAIL');
    depGraph.setProperty('common#Local_Network',
			 'common#consecFailuresFromUser',
			 curTime,
			 0);
    depGraph.setProperty('common#Firefox_Error_Test',
			 'common#ffoxErrorCode',
			 curTime,
			 30);
    var explanations = this.inferRootCause(depGraph);
    correctProbs = {l: 1256, w: 540, d: 9718, i: 1296};
    this.verifyProbs(correctProbs, explanations);

    depGraph = this.initDependencyGraph();
    depGraph.setProperty('common#HTTP_Connection',
			 'core#status',
			 curTime,
			 'FAIL');
    depGraph.setProperty('common#Local_Network',
			 'common#consecFailuresFromUser',
			 curTime,
			 1);
    depGraph.setProperty('common#Firefox_Error_Test',
			 'common#ffoxErrorCode',
			 curTime,
			 30);
    var explanations = this.inferRootCause(depGraph);
    correctProbs = {l: 6717, w: 515, d: 9270, i: 1236};
    this.verifyProbs(correctProbs, explanations);

    depGraph = this.initDependencyGraph();
    depGraph.setProperty('common#HTTP_Connection',
			 'core#status',
			 curTime,
			 'FAIL');
    depGraph.setProperty('common#Local_Network',
			 'common#consecFailuresFromUser',
			 curTime,
			 2);
    depGraph.setProperty('common#Firefox_Error_Test',
			 'common#ffoxErrorCode',
			 curTime,
			 30);
    var explanations = this.inferRootCause(depGraph);
    correctProbs = {l: 8451, w: 507, d: 9127, i: 1217};
    this.verifyProbs(correctProbs, explanations);
    ///////////////////////////////////////////////
    depGraph = this.initDependencyGraph();
    depGraph.setProperty('common#HTTP_Connection',
			 'core#status',
			 curTime,
			 'FAIL');
    depGraph.setProperty('common#Local_Network',
			 'common#consecFailuresFromUser',
			 curTime,
			 0);
    depGraph.setProperty('common#Firefox_Error_Test',
			 'common#ffoxErrorCode',
			 curTime,
			 14);
    var explanations = this.inferRootCause(depGraph);
    correctProbs = {l: 4213, w: 1812, d: 725, i: 4348};
    this.verifyProbs(correctProbs, explanations);

    depGraph = this.initDependencyGraph();
    depGraph.setProperty('common#HTTP_Connection',
			 'core#status',
			 curTime,
			 'FAIL');
    depGraph.setProperty('common#Local_Network',
			 'common#consecFailuresFromUser',
			 curTime,
			 1);
    depGraph.setProperty('common#Firefox_Error_Test',
			 'common#ffoxErrorCode',
			 curTime,
			 14);
    var explanations = this.inferRootCause(depGraph);
    correctProbs = {l: 9121, w: 699, d: 280, i: 1678};
    this.verifyProbs(correctProbs, explanations);

    depGraph = this.initDependencyGraph();
    depGraph.setProperty('common#HTTP_Connection',
			 'core#status',
			 curTime,
			 'FAIL');
    depGraph.setProperty('common#Local_Network',
			 'common#consecFailuresFromUser',
			 curTime,
			 2);
    depGraph.setProperty('common#Firefox_Error_Test',
			 'common#ffoxErrorCode',
			 curTime,
			 14);
    var explanations = this.inferRootCause(depGraph);
    correctProbs = {l: 9651, w: 579, d: 232, i: 1390};
    this.verifyProbs(correctProbs, explanations);


    //////////////////////////////////

    depGraph = this.initDependencyGraph();
    this.dump("with no evidence:");
    var explanations = this.inferRootCause(depGraph);
    correctProbs = {l: 2000, w: 500, d: 1000, i: 1200};
    this.verifyProbs(correctProbs, explanations);
    
    correctProbs = {l: 5587, w: 1397, d: 559, i: 3352};
    this.dump("with ffox evidence:");
    depGraph.setProperty('common#Firefox_Error_Test',
			 'common#ffoxErrorCode',
			 curTime,
			 14);
    var explanations = this.inferRootCause(depGraph);
    this.verifyProbs(correctProbs, explanations);

    this.dump("with failure evidence:");
    depGraph.setProperty('common#HTTP_Connection',
			 'core#status',
			 curTime,
			 'FAIL');
    var explanations = this.inferRootCause(depGraph);
    correctProbs = {l: 5804, w: 1451, d: 580, i: 3483};
    this.verifyProbs(correctProbs, explanations);

    this.dump('with cffu evidence');
    depGraph.setProperty('common#Local_Network',
			 'common#consecFailuresFromUser',
			 curTime,
			 1);
    var explanations = this.inferRootCause(depGraph);
    correctProbs = {l: 9121, w: 699, d: 280, i: 1678};
    this.verifyProbs(correctProbs, explanations);

    var curTime = this.getCurrentTime();

    depGraph.setProperty('common#Firefox_Error_Test',
			 'common#destHash',
			 curTime,
			 'foo');
    depGraph.setProperty('common#Local_Network',
			 'common#consecFailuresFromUser',
			 curTime,
			 1);

    this.dump('Updating beliefs');
    var bel = this.inferRootCause(depGraph);

    if (bel[0].cause == 'l' && Math.round(bel[0].prob * 10000) == 9121) {
      this.dump('inferenceTest 1/3: passed');
    } else {
      this.dump('inferenceTest 1/3: FAILED:' +
		'Pr(' + bel[0].cause + ') == ' + bel[0].prob);
    }

    // this.dump('Setting evidence obconn = FAIL');
    depGraph.setProperty('common#Outbound_Conn_Test',
			 'common#probeResult',
			 curTime,
			 'FAIL');

    // this.dump('Updating beliefs');
    bel = this.inferRootCause(depGraph);
    if (bel[0].cause == 'l' && Math.round(bel[0].prob * 10000) == 9981) {
      this.dump('inferenceTest 2/3: passed');
    } else {
      this.dump('inferenceTest 2/3: FAILED: ' +
		'Pr(' + bel[0].cause + ') == ' + bel[0].prob);
    }

    // this.dump('Setting evidence obconn = FAIL');
    curTime = this.getCurrentTime();
    this.updateConsecFailuresFromUser('x',   curTime, 'OK',   depGraph);
    this.updateConsecFailuresFromUser('foo', curTime, 'FAIL', depGraph);
    this.updateConsecFailuresFromUser('bar', curTime, 'FAIL', depGraph);
    var cffu = this.getConsecFailuresFromUser();

    if (cffu != 2) {
      this.dump('CFFU incorrect: ' + cffu);
      return;
    }
    
    // this.dump('Updating beliefs');
    bel = this.inferRootCause(depGraph);
    if (bel[0].cause == 'l' && Math.round(bel[0].prob * 10000) == 9993) {
      this.dump('inferenceTest 3/3: passed');
    } else {
      this.dump('inferenceTest 3/3: FAILED: ' +
		'Pr(' + bel[0].cause + ') == ' + bel[0].prob);
    }

    
  },

  setCIBEntry: function(obsTime,
			className,
			individualId,
			propertyName,
			value)
  {
    if (value == undefined || value == null) {
      // don't store unitialized values
      return;
    }
    
    // Store data in the component information base.  Overwrite any
    // old data about the property for the component specified.
    //
    if (!(className in this.componentInformationBase)) {
      this.componentInformationBase[className] = {};
    }
    
    if (!(individualId in this.componentInformationBase[className])) {
      this.componentInformationBase[className][individualId] = {};
    }

    this.componentInformationBase[className][individualId][propertyName] = {
      time:  obsTime,
      value: value,
    };
    
    this.clearOldCIBEntries();
  },

  clearOldCIBEntries: function() {
    // clear old entries
    var cibSize = 0;
    var curTime = this.getCurrentTime();
    for (var cibKey in this.componentInformationBase) {
      var numIndividuals = 0;
      // this.dump('clearOldCIBEntries: className ' + cibKey);
      for (var individualId in this.componentInformationBase[cibKey]) {
	var numProps = 0;
	// this.dump('clearOldCIBEntries: individualId ' + individualId);
	for (var propName in this.componentInformationBase[cibKey][individualId]) {
	  // this.dump('clearOldCIBEntries: prop ' + propName);
	  cibSize++;
	  cibTime = this.componentInformationBase[cibKey][individualId][propName].time;
	  if (cibTime < curTime - CIB_LIFETIME) {
	    // expire old CIB entries
	    this.dump('clearOldCIBEntries: expiring ' + individualId + ', ' +
		      propName + ' with time ' + cibTime);
	    delete this.componentInformationBase[cibKey][individualId][propName];
	  } else {
	    numProps++;
	  }
	}
	if (numProps == 0) {
	  delete this.componentInformationBase[cibKey][individualId];
	} else {
	  numIndividuals++;
	}
      }
      if (numIndividuals == 0) {
	delete this.componentInformationBase[cibKey];
      }
    }
    this.dump('CIB size: ' + cibSize);
  },

  getCIBEntry: function(className, individualId, propertyName) {
    // Return a cib entry for a property.  The entry has two fields,
    // value and time.
    //
    var cibClassEntry = this.componentInformationBase[className];
    if (cibClassEntry != null) {
      var cibIndividualEntry = cibClassEntry[individualId];
      if (cibIndividualEntry != null) {
	var val = cibIndividualEntry[propertyName];
	return val;
      }
    }
    return null;
  },

  insertEvidenceFromCIB: function(depGraph, destHash) {
    // add evidence about desthash from the CIB into depGraph
    // TODO: time out the data or something; don't use data that's too old

    // where to find each piece of evidence in the CIB.  The key names
    // are unused and irrelevant right now.  Note that if a property
    // does not appear on this list, it will not get inserted into the
    // depgraph (e.g. hostname).
    var evidenceMap = {
      cffu: {
	subjectName:  'common#Local_Network',
	individualId: SRC_IP,
	propertyName: 'common#consecFailuresFromUser',
      },
      http: {
	subjectName:  'common#HTTP_Connection',
	individualId: destHash,
	propertyName: 'core#status',
      },
      httpTime: {
	subjectName:  'common#HTTP_Connection',
	individualId: destHash,
	propertyName: 'common#connTime',
      },
      httpElapsedTime: {
	subjectName:  'common#HTTP_Connection',
	individualId: destHash,
	propertyName: 'common#elapsedTime',
      },
      destIP: {
	subjectName:  'common#IP_Routing',
	individualId: destHash,
	propertyName: 'common#destIP',
      },
      dnsHostname: {
	subjectName:  'common#DNS_Lookup',
	individualId: destHash,
	propertyName: 'common#hostname',
      },
      dnsIPAddrs: {
	subjectName:  'common#DNS_Lookup',
	individualId: destHash,
	propertyName: 'common#ipAddrs',
      },
      obconn: {
	subjectName:  'common#Outbound_Conn_Test',
	individualId: SRC_IP,
	propertyName: 'common#probeResult',
      },
      obconnURI: {
	subjectName:  'common#Outbound_Conn_Test',
	individualId: SRC_IP,
	propertyName: 'common#probeURI',
      },
      ffox: {
	subjectName:  'common#Firefox_Error_Test',
	individualId: destHash,
	propertyName: 'common#ffoxErrorCode',
      },
      ffoxHash: {
	subjectName:  'common#Firefox_Error_Test',
	individualId: destHash,
	propertyName: 'common#destHash',
      },
      cfts: {
	subjectName:  'common#HTTP_Server',
	individualId: destHash,
	propertyName: 'common#consecFailuresToServer',
      },
    };
	
    for (var evidenceName in evidenceMap) {
      var mapEntry = evidenceMap[evidenceName];
      var entry = this.getCIBEntry(PREFIX + mapEntry.subjectName,
				   mapEntry.individualId,
				   PREFIX + mapEntry.propertyName);
      if (entry != undefined && entry != null) {
	this.dump('Adding evidence: ' +
		  mapEntry.individualId + ': ' +
		  mapEntry.subjectName + ', ' +
		  mapEntry.propertyName + ', ' +
		  entry.time + ', ' +
		  entry.value);
	depGraph.setProperty(mapEntry.subjectName,
			     mapEntry.propertyName,
			     entry.time,
			     entry.value);
      } else {
// 	this.dump('No evidence found for ' +
// 		  mapEntry.subjectName + ', ' +
// 		  mapEntry.individualId + ', ' +
// 		  mapEntry.propertyName);
      }
    }
  },

  extractEvidenceFromDepGraph: function(depGraph) {
    // get the latest evidence about a failure and return it in an
    // associative array.  By default get evidence for the most recent
    // failure.

    if (depGraph == null) {
      this.dump("Warning, no failure detected");
      return;
    }

    // where to find each piece of evidence in the depgraph:
    var evidenceMap = {
      cffu: {
	subjectName:  'common#Local_Network',
	propertyName: 'common#consecFailuresFromUser',
      },
      http: {
	subjectName:  'common#HTTP_Connection',
	propertyName: 'core#status',
      },
      obconn: {
	subjectName:  'common#Outbound_Conn_Test',
	propertyName: 'common#probeResult',
      },
      ffox: {
	subjectName:  'common#Firefox_Error_Test',
	propertyName: 'common#ffoxErrorCode',
      },
      cfts: {
	subjectName:  'common#HTTP_Server',
	propertyName: 'common#consecFailuresToServer',
      },
    };
	
    var evidence = {};

    for (var evidenceName in evidenceMap) {
      var prop = depGraph.getProperty(evidenceMap[evidenceName].subjectName,
				      evidenceMap[evidenceName].propertyName);
      if (prop != null) {
	evidence[evidenceName] = prop.value;
	//	this.dump("Adding evidence: evidence['" + evidenceName +
	//  "'] = " + value);
      } else {
	//	this.dump("depGraph['" + evidenceMap[evidenceName].subjectName +
	//  "'].properties['" + evidenceMap[evidenceName].propertyName +
	//  "'] not found.");
      }
    }

    return evidence;
  },

  initDependencyGraph: function() {
    // construct an initial dependency graph without any evidence
    // filled in yet.
    //

    var depGraph = {
      setProperty: function(className, propertyName, time, value) {
	if (value == undefined || value == null) {
	  // don't set uninitialized or null values
	  return;
	}
	// assumes that there is only one individual for each class name
	this[className].properties[propertyName] = {time: time,
						    value: value};
      },
      
      getProperty: function(className, propertyName) {
	// assumes that there is only one individual for each class name
	var classEntry = this[className];
	if (classEntry == undefined) {
	  return null;
	} else {
	  var prop = this[className].properties[propertyName];
	  return prop;
	}
      },
      
      'common#HTTP_Connection':
      {
	nodeName: 'component',
	id:       0,
	properties:
	{
	  'common#srcIP':      {value: SRC_IP},
	},
	relationships:
	{
	  'common#dnsLookup':  1,
	  'common#ipRouting':  2,
	  'common#dnsLookup':  3,
	  'common#ipRouting':  4,
	},
      },
      'common#DNS_Lookup':
      {
	nodeName: 'component',
	id:       1,
	properties:
	{
	  'common#srcIP':      {value: SRC_IP},
	},
      },
      'common#IP_Routing':
      {
	nodeName: 'component',
	id:       2,
	properties:
	{
	  'common#srcIP':      {value: SRC_IP},
	},
      },
      'common#HTTP_Server':
      {
	nodeName: 'component',
	id:       3,
	properties: {},
      },
      'common#Local_Network':
      {
	nodeName: 'component',
	id:       4,
	properties:
	{
	  'common#ipAddr':   {value: SRC_IP},
	},
      },
      'common#Outbound_Conn_Test':
      {
	nodeName: 'test',
	id:       5,
	properties:
	{
	  'common#srcIP':    {value: SRC_IP},
	},
      },
      'common#Firefox_Error_Test':
      {
	nodeName: 'test',
	properties:
	{
	  'common#srcIP':    {value: SRC_IP},
	},
	relationships:
	{
	  'common#dnsLookup':  1,
	  'common#ipRouting':  2,
	  'common#dnsLookup':  3,
	  'common#ipRouting':  4,
	},
      },
    };
    return depGraph;
  },

  serializeDependencyGraph: function(depGraph) {
    // extract all the evidence and component relationships out of a
    // failure dependency graph and serialize it into XML form.  This
    // lets an agent communicate info in its dependency graph to
    // another agent for recursive diagnosis.
    //
    // in other words, pull out all the relevant info from the
    // component information base
    //
    
    // TODO: take into account evidence about DNS lookups
    // note, order matters because of the relationships pointers

    // TODO: associate times with properties, not observations?

    // For each of the relevant components, get its properties.
    //

    var rootElement = 'observations';
    var doc = this.createXMLDoc(rootElement);
    
    for (var componentName in depGraph) {
      var component = depGraph[componentName];
      var componentNode = doc.createElement(component.nodeName);
      componentNode.setAttribute('id', component.id);
      
      var classNode = doc.createElement('class');
      var classValue = PREFIX + componentName;
      classNode.textContent = classValue;
      componentNode.appendChild(classNode);
      
      for (var propName in component.properties) {
	// TODO: figure out xml namespaces
	// var propNode = doc.createElement(PREFIX + propName);
	var nodeName = PREFIX + propName;
	var parts = nodeName.split('#');
	var propNode = doc.createElement(parts[1]);
	
	var propValue = component.properties[propName];
	if (propValue != null) {
	  propNode.textContent = propValue;
	  componentNode.appendChild(propNode);
	}
      }
      
      for (var relationshipName in component.relationships) {
	// TODO: fix this namespace issue
	var nodeName = PREFIX + relationshipName;
	var parts = nodeName.split('#');

	var relationshipNode = doc.createElement(parts[1]);
	var relationshipValue = component.relationships[relationshipName];
	var refNode = doc.createElement('componentRef');
	refNode.setAttribute('id', relationshipValue);
	relationshipNode.appendChild(refNode);
	componentNode.appendChild(relationshipNode);
      }
    }

    var xml = this.serializeXMLDoc(doc);
    return xml;
  },

  inferRootCause: function(dependencyGraph) {
    // infer the root cause of failure given a probabilistic
    // dependency graph (i.e. bayes net), and evidence.
    //
    // Return a list of possible explanations in order of likelihood
    //
    // for the purposes of this extension, the types of evidence are
    // as follows:
    // 1. whether the host is dead or not (always available)
    // 2. whether a probe succeeded or not (sometimes available)
    // 3. the number of consecutive failures from this user (always available)
    // 4. firefox failure status

    // build a bayes net evidence
    var bnet = this.buildBNetFromDepGraph(dependencyGraph);
    
    // update beliefs in the bayes net
    this.updateBeliefs(bnet);

    // log them for debugging purposes
    // this.logBeliefs(dependencyGraph);
    
    // determine which explanation (cause) is most likely
    var cause = bnet['cause'];
    if (cause == undefined) {
      this.dump("BNet does not contain 'cause'!");
    }

    // must marginalize to just four cases: lWDI, LwDI, LWdI, and LWDi

    var shortCauses = ['l', 'w', 'd', 'i'];
    var probs = [0, 0, 0, 0];
    for (var valIndex=0; valIndex<cause.range.length; valIndex++) {
      var val = cause.range[valIndex];
      for (var statusIndex=0; statusIndex<val.length; statusIndex++) {
	var statusChar = val[statusIndex];
	if (statusChar == statusChar.toLowerCase()) {
	  probs[statusIndex] += cause.bel[val];
	}
      }
    }

    var explanations = new Array();
    for (var causesIndex=0; causesIndex<shortCauses.length; causesIndex++) {
      var shortCause = shortCauses[causesIndex];
      var prob = probs[causesIndex];
      explanations.push( {cause: shortCause, prob: prob} );
      // this.dump('Marginalized cause: Pr(' + shortCause + ') == ' + prob);
    }

    function beliefComparison(a, b) {
      // sort in descending order of beliefs.
      if (a.prob < b.prob) {
	return 1;
      } else if (a.prob == b.prob) {
	return 0;
      } else if (a.prob > b.prob) {
	return -1;
      }
      // should never get here.
    }

    explanations.sort(beliefComparison);

    //     var rv = '';
    //     for (var explIndex=0; explIndex<explanations.length; explIndex++) {
    //       var expl = explanations[explIndex];
    //       rv += "Pr(" + expl + ") == " + cause.bel[expl] + "\n";
    //     }
    // this.dump(rv);

    if (explanations[0].prob > 0) {
      return explanations;
    } else {
      this.dump("uh-oh, inferRootCause returned a weird explanation: " +
		explanations[0].prob);
      return [];
    }
  },

  logLocalDiagnosis: function(dependencyGraph, explanations) {
    // log a diagnosis for evaluation purposes
    // evidence: an associative array of evidence: evidence[nodeName] = value
    // likelihoods[explanation] = prob expl is true
    // explanations: an array of explanations in order of decreasing likelihood

    // what do I need to log?  evidence and the most likely explanation for one thing.

    var diagObs = this.getFailureObservationsFromDepGraph(dependencyGraph);
    diagObs.diagResult = {
      value: explanations[0].cause + ',' + explanations[0].prob,
      time: this.getCurrentTime(),
    }
    if (diagObs.destHash == null) {
      this.dump("Strange, destHash is null: " + diagObs.toSource());
    }

    this.diagHist.push(diagObs);
    this.writeDiagHist();
  },

  makeLocalDiagnosis: function(depGraph) {
    // Diagnose the failure using evidence (or local observations if
    // no evidence is available).
    //
    // Return whether it was able to adequately diagnose the failure
    //
    // Compute the value of each known parent in the CPT
    // Enumerate all valid combinations of parents.
    // Look up probabilities in the CPT and add them up.
    // Return likelihoods.
    this.dump("Performing local diagnosis.");

    var explanations = this.inferRootCause(depGraph);

    if (explanations.length > 0) {
      if (this.lastFailure != null) {
	this.displayDiag(this.lastFailure.host, explanations[0], 'Local agent');
      }
      this.printExplanations(explanations);
      var logging = this.prefs.getBoolPref("extensions.diagtool.logging");
      if (logging) {
	this.logLocalDiagnosis(depGraph, explanations);
      }
    } else {
      this.setToolbarDiagResult('Unknown');
    }
    
    return explanations;
  },

  printExplanations: function(explanations) {
    var explIndex = 0;
    var maxExpls = 4;

    while ((explIndex < explanations.length) && (explIndex < maxExpls)) {
      var expl = explanations[explIndex];
      this.dump("Pr(" + expl.cause + ") == " + expl.prob);
      explIndex++;
    }
  },

  getLastDiagnosis: function() {
    // return the last diagnosis, or null if there was no last diagnosis.
    var diagHist = this.diagHist;
    if (diagHist.length > 0) {
      var diagData = diagHist[diagHist.length - 1];
      return diagData;
    } else {
      return null;
    }
  },

  requestServStats: function(dest) {
    // Get stats on a destination.
    //
    if (dest == null || dest == "") {
      dest = this.currentHost;
    }

    if (dest == null || dest == "") {
      // no host available.
      return;
    }

    if (dest == 'localhost') {
      // As a UI matter, I should gray out the button in this case.
      this.alert("Sorry, server stats are not available for localhost.");
      return;
    }

    // Otherwise request stats/diagnosis from another diagnostic agent
    this.sendStatsRequest(dest);
  },

  requestDiagnosis: function(dest) {
    // This is the function that gets called when the user presses the
    // "Diagnose Failure" button.
    //
    // Diagnose a failure and display the results.  First try a local
    // diagnosis.  If it succeeds, display it and return.  Otherwise,
    // request remote diagnosis.
    //
    // Right now it assumes that dest has failed and that there are
    // corresponding entries in the CIB.
    //
    // TODO: The problem here is that the latest (remote) diaghist
    // entry may wipe out the local entry... so I need to be more
    // careful about that.  Make sure that the remote diagnosis is
    // always at least as accurate as the local one.  i.e. make sure
    // the remote diagnosis is done with the same info as the local
    // one (using the deadhosts).
    //
    // The call tree of requestDiagnosis() is as follows:
    //
    // requestDiagnosis(dest)
    //   initDependencyGraph()
    //   insertEvidenceFromCIB(depGraph, destHash)
    //   makeLocalDiagnosis(depGraph)
    //     inferRootCause(depGraph)
    //   displayDiag(dest, explanations) and return if local diag, else:
    //   sendDiagRequest(dest, depGraph)
    //     createDiagRequest(destHash, depGraph)
    //       getFailureObservationsFromDepGraph(depGraph)
    //       serializeFailureObservation(componentId, failureObservations)
    //     sendMessage(diagAgentURI, text, onLoad, onError)
    //
    // then asynchronously:
    //
    //   receiveDiagResponse(aXMLReq, dest, success)
    //     parseObservations(doc)
    //     parseKnowledge(doc)
    //     parseDiagExplanations(doc)
    //     displayDiag(dest, explanations)
    //

    if (this.lastFailure == null) {
      // Make sure the "Collect Stats" option is checked
      this.alert('No recent failures to diagnose.');
      return;
    }

    // use the latest failure
    if (dest == null || dest == "") {
      dest = this.lastFailure.host;
    }

    if (dest == 'localhost') {
      // As a UI matter, I should gray out the button in this case.
      this.alert("Sorry, diagnosis is not available for localhost.");
      return;
    }

    var depGraph = this.initDependencyGraph();
    var destHash = this.md5sum(dest);
    this.insertEvidenceFromCIB(depGraph, destHash);

    // First try to do a local diagnosis.
    var explanations = this.makeLocalDiagnosis(depGraph);
    if (explanations.length > 0) {
      var explanation = explanations[0];
      var confidence = explanation.prob;

      // var diagData = this.getLastDiagnosis();

      if (confidence >= this.confThresh) {
	this.dump("requestDiagnosis: local diag");
	this.displayDiag(dest, explanations[0], 'Local agent');
	return;
      }
    }
    
    // Otherwise request diagnosis from another diagnostic agent
    this.dump("requestDiagnosis: remote diag");
    this.sendDiagRequest(dest, depGraph);
  },
  
  sendMessage: function(agentURI, reqBody, onLoad, onError) {
    // onload and onerror are callbacks for their respective events.
    // send an xml message to an agentURI
    var request = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Components.interfaces.nsIXMLHttpRequest);

    if (agentURI == null || agentURI == "") {
      this.dump("sendMessage: empty agentURI, aborting");
      return;
    }

    // TODO: need to handle timeouts
    // request.timeout = Components.classes["@mozilla.org/timer;1"].createInstance(Components.interfaces.nsITimer);
    request.overrideMimeType("text/xml");

    var agent = this;
    
    request.onerror = onError;
    request.onload  = onLoad;

    request.open("POST", agentURI, true);  // asynchronous
    this.dump("Sending message to " + agentURI + " with body " + reqBody);

    try {
      request.send(reqBody);
    } catch (e) {
      this.dump("sendMessage: Error sending message.");
    }
  },    

  
  sendConnHistNotification: function() {
    // send a history notification to a server history agent, for example.
    // must make sure we have a histagentmap
    this.dump("Starting conn history notification");

    // TODO must make this time out
    var agent = this;
    var onLoad = function() {
      agent.receiveHistoryNotificationResponse(this, 1);
    };
    var onError = function() {
      agent.receiveHistoryNotificationResponse(this, 0);
    };
    
    var notifications = this.connHistToNotifications();

    // update the number of unacknowledged diagHist entries sent
    this.connHist.entriesSent = this.connHist.entries.length;

    var regionalAgentURI = this.prefs.getCharPref("extensions.diagtool.regionalAgentURI");

    if (notifications.length > 0) {
      var capriMessage = this.constructCapriMessage(notifications);
      this.sendMessage(regionalAgentURI, capriMessage, onLoad, onError);
    }
  },

  sendDiagHistNotification: function() {
    // send a diag history notification to a regional agent.  Only do
    // this after a success so that totalCffu can be set correctly.
    this.dump("Starting diag history notification");

    if (this.diagHist.length == 0) {
      this.dump("No diagHist entries found, aborting.");
      return;
    }

    var notifications = this.diagHistToNotifications();

    // update the number of unacknowledged diagHist entries sent
    this.diagHistEntriesSent = this.diagHist.length;

    var agentURI = this.prefs.getCharPref("extensions.diagtool.regionalAgentURI");

    // TODO must make this time out
    var agent = this;
    var onLoad = function() {
      agent.receiveDiagHistNotificationResponse(this, 1);
    };
    var onError = function() {
      agent.receiveDiagHistNotificationResponse(this, 0);
    };
    
    var capriMessage = this.constructCapriMessage(notifications);
    this.sendMessage(agentURI, capriMessage, onLoad, onError);
  },

  sendKnowledgeRequest: function() {
    // Request probabilistic dependency knowledge from the regional
    // agent.
    var messages = [this.constructMessage('knowledgeRequest',
					  'knowledge:http', '')];

    var agentURI = this.prefs.getCharPref("extensions.diagtool.regionalAgentURI");
    if (agentURI == '') {
      // No agent set yet.  Wait until we have a regional agent to
      // request knowledge from.
      this.dump("sendKnowledgeRequest: no agentURI, aborting");
      return;
    }

    this.dump("sendKnowledgeRequest: starting knowledge request to " +
	      agentURI);
    
    var agent = this;
    var onLoad = function() {
      agent.receiveKnowledgeResponse(this, 1);
    };
    var onError = function() {
      agent.receiveKnowledgeResponse(this, 0);
    };
    
    var capriMessage = this.constructCapriMessage(messages);
    this.sendMessage(agentURI, capriMessage, onLoad, onError);
  },

  sendDiagRequest: function(dest, depGraph) {
    // request diagnosis from a diagnostic agent and display the
    // response.
    //

    var destHash = this.md5sum(dest);
    this.dump('sendDiagRequest, dest ' + dest + ', destHash ' + destHash);
    var messages = this.createDiagRequest(destHash, depGraph);

    var diagAgentURI = this.prefs.getCharPref("extensions.diagtool.regionalAgentURI");
    
    var agent = this;
    var onLoad = function() {
      agent.receiveDiagResponse(this, dest, 1);
    };
    var onError = function() {
      agent.receiveDiagResponse(this, dest, 0);
    };

    this.setToolbarDiagState('Performing remote diagnosis', false);
    var capriMessage = this.constructCapriMessage(messages);
    this.sendMessage(diagAgentURI, capriMessage, onLoad, onError);
  },

  sendStatsRequest: function(dest) {
    // request stats from a stats agent about destHash.
    //

    // First serialize and send connection history info.

    var destHash = this.md5sum(dest);
    this.dump("Starting stats request for dest " + dest +
	      ", desthash " + destHash);

    var text = this.createStatsRequest(destHash);

    var agentURI = this.prefs.getCharPref("extensions.diagtool.regionalAgentURI");
    
    var agent = this;
    var onLoad = function() {
      agent.receiveStatsResponse(this, dest, 1);
    };
    var onError = function() {
      agent.receiveStatsResponse(this, dest, 0);
    };
    
    var capriMessage = this.constructCapriMessage([text]);
    this.sendMessage(agentURI, capriMessage, onLoad, onError);
  },

  probeHosts: function(hosts, depGraph) {
    // Probe the hosts in the hosts array.  This probe is for the
    // failure described by entry.
    //
    function probeCallback(host, status, dg) {
      // create a closure for host and status when the callback is made
      return function() {
	  agent.probeResponse(this, host, status, dg);
      };
    }

    var agent = this;
    for (var i=0; i<hosts.length; i++) {
      var request = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Components.interfaces.nsIXMLHttpRequest);
      // TODO: need to handle timeouts
      // request.timeout = Components.classes["@mozilla.org/timer;1"].createInstance(Components.interfaces.nsITimer);

      // 128 for successful probe, 129 for failed probe
      var onerror = probeCallback(hosts[i], 129, depGraph);
      var onload  = probeCallback(hosts[i], 128, depGraph);
      
      request.addEventListener("load", onload, false);
      request.addEventListener("error", onerror, false);

      request.open("GET", hosts[i], true);  // asynchronous
      this.dump("Sending probe to " + hosts[i]);
      this.setToolbarDiagState('Testing local network', false);
      try {
        request.send(null);
      }
      catch (e) {
	this.dump("Error sending probe to " + hosts[i]);
      }
    
    }
  },

  getProbeResult: function(errorCode) {
    // translate from a numeric errorCode to a text result
    var result;
    if (errorCode == 128) {
      result = 'OK';
    } else if (errorCode == 129) {
      result = 'FAIL';
    }
    return result;
  },

  getCurrentTime: function() {
    var d = new Date();
    var curTime = d.getTime();
    return curTime;
  },
    
  probeResponse: function(aXMLReq, uri, errorCode, depGraph) {
    // update the depGraph and CIB to reflect information from the
    // probe

    this.dump("Probe result: " + uri + ", " + errorCode);
    this.setToolbarDiagState('', true);
    var host = this.extractHost(uri);
    var curTime = this.getCurrentTime();
    var result = this.getProbeResult(errorCode);
    this.updateConnHist({
      host: host,
	  connTime: curTime,
	  errorCode: errorCode,
	  status: result // status is same as result
	  }
      );
    

    // update consecfailuresfromuser
    this.updateConsecFailuresFromUser(host, curTime, result, depGraph)

    // add the new evidence to the failure dependency graph
    depGraph.setProperty('common#Outbound_Conn_Test',
			 'common#probeResult',
			 curTime,
			 result);

    depGraph.setProperty('common#Outbound_Conn_Test',
			 'common#probeURI',
			 curTime,
			 uri);

    this.setCIBEntry(curTime,
		     PREFIX + 'common#Outbound_Conn_Test',
		     SRC_IP,
		     PREFIX + 'common#probeResult',
		     result);

    this.setCIBEntry(curTime,
		     PREFIX + 'common#Outbound_Conn_Test',
		     SRC_IP,
		     PREFIX + 'common#probeURI',
		     host);

    // Make a local diagnosis using the new evidence from the probe
    this.makeLocalDiagnosis(depGraph);

    if (this.prefs.getBoolPref('extensions.diagtool.autoDiag')) {
      var destHost = depGraph.getProperty('common#DNS_Lookup',
					  'common#hostname')['value'];
      this.sendDiagRequest(destHost, depGraph);
    }
  },
    
  getDateParts: function(t) {
    var d = new Date();
    var curTime = d.setTime(t);
    var rv = {};
    rv.month = d.getMonth() + 1;
    rv.date = d.getDate();
    rv.year = d.getFullYear();
    rv.hour = d.getHours();
    var min = d.getMinutes();
    min = ("0" + min.toString()).substr(-2);
    rv.min = min;
    return rv;
  },
    
  dateToString: function(t) {
    var d = new Date();
    d.setTime(t);
    var dateParts = this.getDateParts(t);
    // e.g. "Wed 19 Jan 2005 01:30:00 AM EST"
    var ds = d.toString().split(' ');
    // e.g. "Jan 19 01:30"
    return [ds[1], ds[2],
	    dateParts.hour.toString() + ':' + dateParts.min].join(' ');
  },

  getConnSummary: function() {
    return this.connSummary;
  },

  // update the connection summary based on the new connection
  updateConnSummary: function(entry) {
    var node = RDF.GetResource(PREFIX + "ua#host-" + entry.host);

    var connsProp          = RDF.GetResource(PREFIX + "ua#conns");
    var failuresProp       = RDF.GetResource(PREFIX + "ua#failures");
    var failureRateProp    = RDF.GetResource(PREFIX + "ua#failureRate");
    var intFailureRateProp = RDF.GetResource(PREFIX + "ua#intFailureRate");
    var lastConnTimeProp   = RDF.GetResource(PREFIX + "ua#lastConnTime");
    var lastConnTimeStringProp = RDF.GetResource(PREFIX + "ua#lastConnTimeString");
    var hostProp         = RDF.GetResource(PREFIX + "ua#host");

    var conns          = this.connSummary.GetTarget(node, connsProp, true);
    var failures       = this.connSummary.GetTarget(node, failuresProp, true);
    var failureRate    = this.connSummary.GetTarget(node, failureRateProp, true);
    var intFailureRate = this.connSummary.GetTarget(node, intFailureRateProp, true);
    var lastConnTime = this.connSummary.GetTarget(node, lastConnTimeProp, true);
    var lastConnTimeString = this.connSummary.GetTarget(node, lastConnTimeStringProp, true);
    var failure = (entry.status == 'OK') ? 0 : 1;

    if (conns instanceof Components.interfaces.nsIRDFInt) {
      // update an existing entry
      var c = parseInt(conns.Value) + 1;
      var newConns = RDF.GetIntLiteral(c);
      this.connSummary.Change(node, connsProp, conns, newConns);
      
      if ((failures       instanceof Components.interfaces.nsIRDFInt) &&
	  (failureRate    instanceof Components.interfaces.nsIRDFLiteral) &&
	  (intFailureRate instanceof Components.interfaces.nsIRDFInt) &&
	  (lastConnTime   instanceof Components.interfaces.nsIRDFInt) &&
	  (lastConnTimeString instanceof Components.interfaces.nsIRDFLiteral)) {
	var f = failures.Value + failure;
	var newFailures = RDF.GetIntLiteral(f);
	var rate = f/c*100;
	var intRate = parseInt(f/c*10000);
	var newFailureRate = RDF.GetLiteral(rate.toFixed(2));
	var newIntFailureRate = RDF.GetIntLiteral(intRate);
	var connTime = entry.connTime;
	var connTimeString = this.dateToString(entry.connTime);

	this.connSummary.Change(node, failuresProp, failures, newFailures);
	this.connSummary.Change(node, failureRateProp, failureRate, newFailureRate);
	this.connSummary.Change(node, intFailureRateProp, intFailureRate, newIntFailureRate);
	this.connSummary.Change(node, lastConnTimeProp, lastConnTime,
				RDF.GetIntLiteral(connTime));
	this.connSummary.Change(node, lastConnTimeStringProp, lastConnTimeString,
				RDF.GetLiteral(connTimeString));
	this.obs.notifyObservers(null, "diagtool-update-history", null);
      }
    } else {
      // create a new entry
      // this.dump("Creating entry with URI " + CONN_SUMMARY_URI);
      var rootnode = RDF.GetResource(CONN_SUMMARY_URI);
      var con = this.getContainer(this.connSummary, rootnode);
      var rate = failure * 100;
      rate = rate.toFixed(2);
      var connTime       = entry.connTime;
      var connTimeString = this.dateToString(entry.connTime);

      this.connSummary.Assert(node, connsProp,
			      RDF.GetIntLiteral(1),            true);
      this.connSummary.Assert(node, hostProp,
			      RDF.GetLiteral(entry.host),      true);
      this.connSummary.Assert(node, lastConnTimeProp,
			      RDF.GetIntLiteral(connTime),     true);
      this.connSummary.Assert(node, lastConnTimeStringProp,
			      RDF.GetLiteral(connTimeString),  true);
      this.connSummary.Assert(node, failuresProp,
			      RDF.GetIntLiteral(failure),      true);
      this.connSummary.Assert(node, failureRateProp,
			      RDF.GetLiteral(rate),            true);
      this.connSummary.Assert(node, intFailureRateProp,
			      RDF.GetIntLiteral(rate*100),     true);
      con.InsertElementAt(node, 1, true);
    }
  },

  resetConnSummary: function() {
    var rootnode = RDF.GetResource(CONN_SUMMARY_URI);
    var con = this.getContainer(this.connSummary, rootnode);
    var elements = con.GetElements();

    while (elements.hasMoreElements()) {
      var element = elements.getNext();
      var preds = this.connSummary.ArcLabelsOut(element);
      while (preds.hasMoreElements()) {
	var pred = preds.getNext();
	var target = this.connSummary.GetTarget(element, pred, true);
	this.connSummary.Unassert(element, pred, target);
      }
      con.RemoveElement(element, false);
    }
  },

  // add a new entry to the connection history
  updateConnHist: function(entry) {
    var logging = this.prefs.getBoolPref("extensions.diagtool.logging");
    if (!logging) {
      return;
    }

    // this.dump("Updating connHist: ");
    if (entry.host == 'localhost') {
      this.dump("Not logging localhost");
      return;
    }
    entry.hostHash = this.md5sum(entry.host);

    var queue = this.connHist.entries;
    queue.push(entry);

    // this.dump("Done updating connHist");
    if (!("interrupted" in entry)) {
      // don't include interrupted connections in the summary
      this.updateConnSummary(entry);
      this.writeConnSummary();
    }

    // this.dump('Inserted entry with hostHash ' + queue[queue.length - 1].hostHash);
    this.writeConnHist();
  },

  // TODO: make it work with wyciwyg URLs
  extractHost: function(aURI) {
      // regular expression from RFC2396 12.B.
      var reg = new RegExp("^(([^:/?#]+):)?(//([^/?#]*))?([^?#]*)(\\?([^#]*))?(#(.*))?");
      var match = reg.exec(aURI);
      if (match == null) {
	this.dump("Could not parse URI");
	return null;
      }

      var hostname = match[4];
      return hostname;
  },

  updateConsecFailuresFromUser: function(hostname, curTime, status, depGraph) {
    // update the number of consecutive failures from a user with a
    // new connection with status dest.  Also update the depgraph and
    // CIB.  Note that status here is 'OK' or 'FAIL'
    //
    if (status == 'OK') {
      this.consecFailedDests = {};
    } else if (status == 'FAIL') {
      this.consecFailedDests[hostname] = 1;
    } else {
      this.dump("Invalid status: " + status);
    }

    var cffu = this.getConsecFailuresFromUser();

    if (depGraph != null) {
      depGraph.setProperty('common#Local_Network',
			   'common#consecFailuresFromUser',
			   curTime,
			   cffu);
    }
    
    this.setCIBEntry(curTime,
		     PREFIX + 'common#Local_Network',
		     SRC_IP,
		     PREFIX + 'common#consecFailuresFromUser',
		     cffu);
  },

  getConsecFailuresFromUser: function() {
    var count = 0;
    for (key in this.consecFailedDests) {
      count++;
    }
    var maxCffu = this.nodeAttributes['cffu'].range.length - 1;
    if (count > maxCffu) {
      count = maxCffu;
    }
    return count;
  },

  addFailureCIBEntries: function(entry) {
    // Adds the failure info in entry to the CIB
    //

    // remove any trailing port numbers from hostname
    var hostnameParts = entry.host.split(':');
    var hostname = hostnameParts[0];
    this.setCIBEntry(entry.connTime,
		     PREFIX + 'common#DNS_Lookup',
		     entry.hostHash,
		     PREFIX + 'common#hostname',
		     hostname);

    this.setCIBEntry(entry.connTime,
		     PREFIX + 'common#DNS_Lookup',
		     entry.hostHash,
		     PREFIX + 'common#ipAddrs',
		     entry.ipAddrs);

    this.setCIBEntry(entry.connTime,
		     PREFIX + 'common#IP_Routing',
		     entry.hostHash,
		     PREFIX + 'common#destIP',
		     entry.destIP);

    this.setCIBEntry(entry.connTime,
		     PREFIX + 'common#Firefox_Error_Test',
		     entry.hostHash,
		     PREFIX + 'common#destHash',
		     entry.hostHash);

    this.setCIBEntry(entry.connTime,
		     PREFIX + 'common#Firefox_Error_Test',
		     entry.hostHash,
		     PREFIX + 'common#ffoxErrorCode',
		     entry.errorCode);

    if (entry.status != 'FAIL') {
      this.dump('entry.status == ' + entry.status + ' != FAIL!');
    }

    this.setCIBEntry(entry.connTime,
		     PREFIX + 'common#HTTP_Connection',
		     entry.hostHash,
		     PREFIX + 'core#status',
		     entry.status);

    this.setCIBEntry(entry.connTime,
		     PREFIX + 'common#HTTP_Connection',
		     entry.hostHash,
		     PREFIX + 'common#connTime',
		     entry.connTime);

    if (entry.elapsedTime != undefined) {
      this.setCIBEntry(entry.connTime,
		       PREFIX + 'common#HTTP_Connection',
		       entry.hostHash,
		       PREFIX + 'common#elapsedTime',
		       entry.elapsedTime);
    }

    this.setCIBEntry(entry.connTime,
		     PREFIX + 'common#HTTP_Connection',
		     entry.hostHash,
		     PREFIX + 'common#destHash',
		     entry.hostHash);

  },

  createFailureDependencyGraph: function(entry) {
    // given a failure entry, return a failure dependency graph that
    // contains the info in the failure entry.
    //
    // entry must contain at least the following keys:
    // connTime, host, hostHash, status
    
    var depGraph = this.initDependencyGraph();
	
    depGraph.setProperty('common#Firefox_Error_Test',
			 'common#destHost',
			 entry.connTime,
			 entry.host);

    depGraph.setProperty('common#Firefox_Error_Test',
			 'common#destHash',
			 entry.connTime,
			 entry.hostHash);

    depGraph.setProperty('common#Firefox_Error_Test',
			 'common#ffoxErrorCode',
			 entry.connTime,
			 entry.errorCode);

    if (entry.status != 'FAIL') {
      this.dump('createFailureDependencyGraph: entry.status == ' +
		 entry.status + ' != FAIL!');
    }
    
    depGraph.setProperty('common#HTTP_Connection',
			 'core#status',
			 entry.connTime,
			 entry.status);

    depGraph.setProperty('common#HTTP_Connection',
			 'common#connTime',
			 entry.connTime,
			 entry.connTime);

    depGraph.setProperty('common#HTTP_Connection',
			 'common#destHash',
			 entry.connTime,
			 entry.hostHash);
	
    depGraph.setProperty('common#DNS_Lookup',
			 'common#hostname',
			 entry.connTime,
			 entry.host);
	
    depGraph.setProperty('common#DNS_Lookup',
			 'common#ipAddrs',
			 entry.connTime,
			 entry.ipAddrs);
	
    depGraph.setProperty('common#IP_Routing',
			 'common#srcIP',
			 entry.connTime,
			 entry.host);

    depGraph.setProperty('common#IP_Routing',
			 'common#destIP',
			 entry.connTime,
			 entry.destIP);
    
    return depGraph;
  },

  logConnection: function(entry) {
    // Record a connection, including logging it and updating
    // consecutive failures from user.
    this.currentHost = entry.host;
    this.updateConnHist(entry);
    this.updateConsecFailuresFromUser(entry.host, entry.connTime,
				      entry.status, null);
  },

  setLastFailure: function(entry) {
    this.lastFailure = entry;
    this.setToolbarLastFailure(entry.host);
  },

  setToolbarDiagState: function(stateValue, diagComplete) {
    // The current state and whether diagnosis is complete
    var data = stateValue + '|' + diagComplete;
    this.obs.notifyObservers(null, "diagtool-set-diag-state", data);
  },

  setToolbarDiagResult: function(result) {
    this.obs.notifyObservers(null, "diagtool-set-diag-result", result);
  },

  setToolbarLastFailure: function(host) {
    this.obs.notifyObservers(null, "diagtool-set-last-failure", host);
  },

  diagnoseFailure: function(entry) {
    // Handle a connection failure, including logging it and
    // diagnosing it if necessary.
    this.setToolbarDiagState('Performing local diagnosis', false);

    // We will make three diagnoses.
    //
    // 1. First, do a local diagnosis without conducting any
    // additional tests.  i.e. using only cffu cfts (if known)
    // and ffoxErrorCode.
    //
    // 2. Do a local diagnosis with obconn data. (after
    // probeHosts() in probeResponse())
    //
    // 3. Do a remote diagnosis (asynchronously after the probehosts
    // response)
    //
    // The idea is that by collecting all this data, I can then
    // compare later on how useful a probe is, how useful remote
    // cfts data is, etc.
    // 
    var depGraph = this.createFailureDependencyGraph(entry);
    this.updateConsecFailuresFromUser(entry.host, entry.connTime,
				      'FAIL', depGraph);
	  
    this.addFailureCIBEntries(entry);

    var destHash = this.md5sum(entry.host);
    this.insertEvidenceFromCIB(depGraph, destHash);
    this.makeLocalDiagnosis(depGraph);

    // when a failure occurs, probe a known good server to check
    // if it's a user failure
    var probing = this.prefs.getBoolPref("extensions.diagtool.baselineProbing");
    if (probing) {
      var probeHost = this.prefs.getCharPref("extensions.diagtool.probeURI");
      this.probeHosts([probeHost], depGraph);
      // asynchronously request remote diagnosis after the probe
    } else if (this.prefs.getBoolPref('extensions.diagtool.autoDiag')) {
      this.sendDiagRequest(entry.host, depGraph);
    } else {
      this.setToolbarDiagState('', true);
    }
  },

  /////////////////////////
  // nsIWebProgressListener
  //
  onStateChange: function(aProgress, aRequest, aFlag, aStatus)
  {
    if (!this.initialized) {
      return 0;
    }

    // only consider network state changes
    if (!(aFlag & Components.interfaces.nsIWebProgressListener.STATE_IS_NETWORK)) {
      return 0;
    }

    if (aRequest == null) {
      // This happens sometimes if the user clicks on a different tab
      this.dump("Warning: aRequest is null.  aFlag: " + aFlag + ", aStatus: " + aStatus);
      return 0;
    }

    var curTime = this.getCurrentTime();

    if (aFlag & STATE_START) {
      // This fires when the load event is initiated
      aRequest = aRequest.QueryInterface(Components.interfaces.nsIRequest);
      var name = aRequest.name;
      this.startTime[name] = curTime;
      var numKeys = 0;
      for (var key in this.startTime) {
	numKeys++;
      }
      this.dump('this.startTime size: ' + numKeys);
      // this.dump("STATE_START URI: " + name);
    }

    if (aFlag & STATE_STOP) {
      // This fires when the load finishes
      aRequest = aRequest.QueryInterface(Components.interfaces.nsIRequest);
      var name = aRequest.name;

      // this.dump("STATE_STOP URI: " + name);

      var hostname = this.extractHost(name);
      if (hostname == null) {
	this.dump("Could not parse URI");
	return 0;
      } else {
	// this.dump("Host: " + hostname);
      }

      var errorCode = aStatus & 0xffff;
      if (errorCode > 255) {
	this.dump("Warning: unknown error code, skipping.  Status: " +
		  aStatus + "Error code: " + errorCode);
	return 0;
      }
	
      var entry = {};

      // see mozilla/netwerk/base/public/nsNetError.h for the meaning
      // of the various error codes
     
      switch (errorCode) {
      case 0: 
	// success.  Ignore this because success will be handled by
	// the http-examine-response observer.
      case 3: // Request redirected
      case 4: // retargeted
      case 8: // I don't know what this one is, but it doesn't seem to be an error.
      case 10: // malformed uri
      case 15: // in progress
      case 16: // offline
      case 17: // no content
      case 18: // unknown protocol
      case 19: // port access not allowed
      case 25: // not resumable
      case 32: // entity changed
      case 31: // redirect loop
	
      case 51: // unknown socket type
      case 52: // socket create failed
	
      case 61: // cache related error codes
      case 62:
      case 63:
      case 64:
      case 65:
      case 66:
      case 67:
      case 68:
      case 70:
	// not network related errors
	this.dump("No network error.  error code: " + errorCode);
	// This is necessary because sometimes the STATE_STOP occurs
	// before the http-on-examine-response if it was preloaded.
	if (name in this.startTime) {
	  delete this.startTime[name];
	}
	break;
      case 128:
      case 129:
	// reserve for probe status
	this.dump("Invalid error code; reserved for probe status.  error code: " + errorCode);
	break;
      case 2:
	// Request canceled (e.g. Stop or refresh button pressed).
	// Not necessarily a failure, but if elapsed time is large
	// then it might indicate poor performance.
        entry.interrupted = 1;
	// FALL THROUGH to default
      case 13: // connection refused
      case 14: // timeout
	
      case 20: // connection reset
      case 21: // ftp login
      case 22: // ftp cwd
      case 23: // ftp pasv
      case 24: // ftp pwd
      case 25: // ftp list
	
      case 30: // dns lookup failed
	
      case 42: // proxy dns lookup failed
	
      case 71: // net interrupt
      case 72: // proxy connection refused
      
      default:
	// do two things: 1. log the connection to send in a
	// connection history, and 2. diagnose it if its a failure.
	entry.host = hostname;
	entry.hostHash = this.md5sum(hostname);
        entry.connTime = curTime;
	entry.errorCode = errorCode;
	
	if (name in this.startTime) {
	  entry.elapsedTime = curTime - this.startTime[name];
	  // only count it once
	  delete this.startTime[name];
	}

	// It's a failure unless it was interrupted before
	// ELAPSED_TIME_FAILURE_THRESH
	//
	var isFailure = (!('interrupted' in entry) ||
			 (('elapsedTime' in entry) &&
			  entry.elapsedTime > ELAPSED_TIME_FAILURE_THRESH));
	
	entry.status = isFailure ? 'FAIL' : 'OK';
	
	// get the IP address of the failed destination, unless it's
	// DNS lookup failure (or an interrupted connection, which may
	// also be a DNS failure)

	// start an asynchronous lookup; when this completes, handle
	// the failure.
	this.dnsLookupAndProcessEntry(entry);
      }
    }
    return 0;
  },

  updateDiagHist: function() {
    // first, find the total cffu, which is the cffu of the most
    // recent failure.  Then set all the other consecutive failures to
    // that cffu.
    //
    // The problem is race conditions due to the probe.  Basically,
    // going from back to front the diaghist.cffu will look something
    // like this:
    // 0 0 1 0 2 1 0 5 4 3 3 2 1 0
    // I want to make it look like this:
    // 0 0 1 0 2 2 0 5 5 5 5 5 5 0
    //
    // There are two states: looking for a cffu, and setting cffu.
    // Looking for a cffu is indicated by a 0 under the current
    // index. (totalcffu = 0)
    var totalCffu = 0;
    for (var diagHistIndex=this.diagHist.length-1;
	 diagHistIndex>=0;
	 diagHistIndex--) {
      var diagHistEntry = this.diagHist[diagHistIndex];
      var currentCffu;
      if (diagHistEntry.cffu != undefined) {
	currentCffu = diagHistEntry.cffu.value;
      } else {
	this.dump("Whoops, couldn't get cffu value!");
	return;
      }

      // if we are in the seeking state, set totalCffu to the next cffu.
      // if the current cffu is 0, then reset totalcffu to the seeking state.
      if (totalCffu == 0 || currentCffu == 0) {
	totalCffu = currentCffu;
      }	

      // set the current totalcffu
      if (diagHistEntry.totalCffu == undefined ||
	  diagHistEntry.totalCffu == null) {
	diagHistEntry.totalCffu = {
	  value: totalCffu,
	  time: this.getCurrentTime()
	};
	if (diagHistEntry.destHash != null) {
	  this.dump("Updating diagHist, setting totalCffu for dest "
		    + diagHistEntry.destHash.value + " = " + totalCffu);
	}
      } else {
	break;
      }
    }
  },

  
  onLocationChange: function(aProgress, aRequest, aURI) {
    // Keep track of the current host for deciding what server to get
    // servstats for.
    this.currentHost = aURI.asciiHost;

    // enable the servstats button
    this.obs.notifyObservers(null, "diagtool-servstats-disabled", "false");
    // this.dump("onLocationChange() URI: " + aURI.spec + ", Host: " + this.currentHost);
  },

  onProgressChange: function() { return 0; },
  onStatusChange: function() { return 0; },
  onSecurityChange: function() {return 0;},
  onLinkIconAvailable: function() {return 0;},

  /////////////////////////////
  // nsIObserver
  observe: function(aSubject, aTopic, aData)
  {
    var enabled, prefix;
    var curTime = this.getCurrentTime();
    
    //do nothing until initialized
    if (!this.initialized)
      return;

    if (aTopic == 'timer-callback') {
      // Periodically check the conn history log and send out notifications
      // Don't let the histsize grow too much

      if (this.connHist.entries.length > MAX_HIST_SIZE) {
	this.sendConnHistNotification();
      }

      if (this.diagHist.length > MAX_DIAG_HIST_SIZE) {
	// the total number of failures for all previous diagHist
	// entries is the number of failures just before a success.
	this.updateDiagHist();
	this.sendDiagHistNotification();
      }
      
      // if our agent map is too old, get a new one.
      var agentMapExpires = 1000 * this.prefs.getIntPref("extensions.diagtool.agentMapExpires");
      if (agentMapExpires < curTime) {
	this.loadAgentMap(true);
      }
    }
    else if (aTopic == "http-on-examine-response") {
      // I don't use EndDocumentLoad here because that also counts
      // cached pages.  http-on-examine response logs all successfully
      // retrieved uncached pages.  I'm going to ignore cached page
      // loads.  errorCode OK is success.
      //
      // Only log successes that are ordinary page loads (i.e. not
      // image loads, xmlhttprequests, etc.
      //
      // Note that sometimes firefox preloads links, and so that the
      // onStateChange occurs after the http-on-examine-response here.
      // For simplicity, we will not count those as successes either.
      //
      aSubject.QueryInterface(Components.interfaces.nsIHttpChannel);
      var hostname = aSubject.URI.asciiHost;
      var uri = aSubject.URI.spec;
      var entry = {
	host: hostname,
	connTime: curTime,
	status: 'OK',
	errorCode: OK
      };
      // this.dump("http-on-examine-response URI: " + uri);

      // Only consider connections that were noted in onStateChange START
      if (uri in this.startTime) {
	entry.elapsedTime = curTime - this.startTime[uri];

	if (this.debug && entry.elapsedTime > 60000) {
	  // I think the reason this happens is that there are times
	  // when the onStateChange START occurs after the actual
	  // http-on-examine-response, and so the next time the page
	  // is loaded much later, the next http-on-examine-response
	  // matches the previous START.
	  //
	  // I think I quashed this bug by removing names from
	  // startTime in STATE_STOP of onStateChange.  Therefore this
	  // message below should never appear.
	  var msg = uri + " elapsed time is " + entry.elapsedTime;
	  this.dump(msg);
	}
	
	// Remove the uri to prevent inaccurate latencies in the
	// future
	delete this.startTime[uri];

	// ignore connections to localhost
	if (hostname != 'localhost') {
	  this.dump("**** SUCCESS: " + hostname);
	  this.dnsLookupAndProcessEntry(entry);
	}
      }
    }
    else if (aTopic == "profile-before-change") {
      this.stop();
      this.initialized = false;
    }
  },
     
  ///////////////////////////
  // nsIClassInfo  
  getInterfaces: function(aCount)
  {
    var ifaces = new Array();
    ifaces.push(Components.interfaces.nsIDiagAgent);
    ifaces.push(Components.interfaces.nsIObserver);
    ifaces.push(Components.interfaces.nsIWebProgressListener);
    ifaces.push(Components.interfaces.nsISupportsWeakReference);
    ifaces.push(Components.interfaces.nsIClassInfo);
    ifaces.push(Components.interfaces.nsISupports);
    aCount.value = ifaces.length;
    return ifaces;
  },
  
  getHelperForLanguage: function(aLanguage) { return null; },
  get contractID() { return CONTRACT_ID; },
  get classID() { return CLASS_ID; },
  get classDescription() { return CLASS_NAME; },
  get implementationLanguage() { return Components.interfaces.nsIProgrammingLanguage.JAVASCRIPT; },
  get flags() { return Components.interfaces.nsIClassInfo.SINGLETON; },
   
  ///////////////////////////
  // nsISupports   
  QueryInterface: function(iid)
  {
    if (!iid.equals(Components.interfaces.nsIDiagAgent) &&
        !iid.equals(Components.interfaces.nsIObserver) &&
        !iid.equals(Components.interfaces.nsIClassInfo) &&
        !iid.equals(Components.interfaces.nsIWebProgressListener) &&
        !iid.equals(Components.interfaces.nsISupportsWeakReference) &&
        !iid.equals(Components.interfaces.nsISupports))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  }
}

/******************************************************************************
 * XPCOM Functions for construction and registration
 ******************************************************************************/
var gModule = {
  initialized: false,
  registerSelf: function(aCompMgr, aFileSpec, aLocation, aType)
  {
    if (!this.initialized) {
      this.initialized = true;
      throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
    };
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(CLASS_ID, CLASS_NAME, CONTRACT_ID, aFileSpec, aLocation, aType);
  },

  unregisterSelf: function(aCompMgr, aLocation, aType)
  {
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.unregisterFactoryLocation(CLASS_ID, aLocation);     
  },
  
  getClassObject: function(aCompMgr, aCID, aIID)
  {
    if (!aIID.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    if (aCID.equals(CLASS_ID))
      return gFactory;

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  canUnload: function(aCompMgr) { return true; }
};

var gFactory = {
  createInstance: function(aOuter, aIID)
  {
    if (aOuter != null)
      throw Components.results.NS_ERROR_NO_AGGREGATION;

    var agent = new diagAgent();
    agent.start();
    return agent.QueryInterface(aIID);
  }
};

function NSGetModule(aCompMgr, aFileSpec) { return gModule; }


//////////////////////////////////////////////////////////////////////
// Belief propagation for polytrees inference algorithm
//
// TODO: recursively propagate/request messages
//////////////////////////////////////////////////////////////////////

function runtests() {
  var testText = document.getElementById("testText");
  var results = bnTest();
  results += inferenceTest();
  testText.innerHTML = results;
}

function inferenceTest() {
  var rv = "";

  var bnet = {};
  var cause = new bnNode();
  cause.bnet = bnet;
  bnet['cause'] = cause;
  cause.name = "cause";
  // lowercase for failed, uppercase for OK
  // Local net, Web server, DNS, IP
  cause.range = [
		 'lwdi',
		 'lwdI',
		 'lwDi',
		 'lwDI',
		 'lWdi',
		 'lWdI',
		 'lWDi',
		 'lWDI',
		 'Lwdi',
		 'LwdI',
		 'LwDi',
		 'LwDI',
		 'LWdi',
		 'LWdI',
		 'LWDi',
		 'LWDI',
		 ];
  cause.children = ['obconn', 'cffu', 'http', 'cfts', 'ffox'];
  cause.cpt = {};
  cause.cpt['lwdi|'] = 1.0 / 16;
  cause.cpt['lwdI|'] = 1.0 / 16;
  cause.cpt['lwDi|'] = 1.0 / 16;
  cause.cpt['lwDI|'] = 1.0 / 16;
  cause.cpt['lWdi|'] = 1.0 / 16;
  cause.cpt['lWdI|'] = 1.0 / 16;
  cause.cpt['lWDi|'] = 1.0 / 16;
  cause.cpt['lWDI|'] = 1.0 / 16;
  cause.cpt['Lwdi|'] = 1.0 / 16;
  cause.cpt['LwdI|'] = 1.0 / 16;
  cause.cpt['LwDi|'] = 1.0 / 16;
  cause.cpt['LwDI|'] = 1.0 / 16;
  cause.cpt['LWdi|'] = 1.0 / 16;
  cause.cpt['LWdI|'] = 1.0 / 16;
  cause.cpt['LWDi|'] = 1.0 / 16;
  cause.cpt['LWDI|'] = 1.0 / 16;

  var obconn = new bnNode();
  obconn.bnet = bnet;
  bnet['obconn'] = obconn;
  obconn.name = 'obconn';
  obconn.range = ['OK', 'FAIL'];
  obconn.parents = ['cause'];
  obconn.cpt = {};
  for (var i=0; i<cause.range.length; i++) {
    var causeVal = cause.range[i];
    if (causeVal[0] == 'l') {
      // local net failed
      obconn.cpt['OK|'   + causeVal] = 0.0;
      obconn.cpt['FAIL|' + causeVal] = 1.0;
    } else {
      // local net okay
      obconn.cpt['OK|'   + causeVal] = 0.98;
      obconn.cpt['FAIL|' + causeVal] = 0.02;
    }
  }

  var cffu = new bnNode();
  cffu.bnet = bnet;
  bnet['cffu'] = cffu;
  cffu.name = 'cffu';
  cffu.range = [0, 1, 2];
  cffu.parents = ['cause'];
  cffu.cpt = {};

  for (var i=0; i<cause.range.length; i++) {
    var causeVal = cause.range[i];
    if (causeVal[0] == 'l') {
      // local net failed
      cffu.cpt['0|' + causeVal] = 0.5;
      cffu.cpt['1|' + causeVal] = 0.3;
      cffu.cpt['2|' + causeVal] = 0.2;
    } else {
      // local net okay
      cffu.cpt['0|' + causeVal] = 0.95;
      cffu.cpt['1|' + causeVal] = 0.04;
      cffu.cpt['2|' + causeVal] = 0.01;
    }
  }

  var http = new bnNode();
  http.bnet = bnet;
  bnet['http'] = http;
  http.name = 'http';
  http.range = ['OK', 'FAIL'];
  http.parents = ['cause'];
  http.cpt = {};

  for (var i=0; i<cause.range.length; i++) {
    var causeVal = cause.range[i];
    if (causeVal == 'LWDI') {
      // everything is okay
      http.cpt['OK|'   + causeVal] = 1.0;
      http.cpt['FAIL|' + causeVal] = 0.0;
    } else {
      // something has failed
      http.cpt['OK|'   + causeVal] = 0.0;
      http.cpt['FAIL|' + causeVal] = 1.0;
    }
  }

  var cfts = new bnNode();
  cfts.bnet = bnet;
  bnet['cfts'] = cfts;
  cfts.name = 'cfts';
  cfts.range = [0, 1, 2];
  cfts.parents = ['cause'];
  cfts.cpt = {};

  for (var i=0; i<cause.range.length; i++) {
    var causeVal = cause.range[i];
    if (causeVal[1] == 'w') {
      // web server failed
      cfts.cpt['0|' + causeVal] = 0.1;
      cfts.cpt['1|' + causeVal] = 0.3;
      cfts.cpt['2|' + causeVal] = 0.6;
    } else {
      // web server okay
      cfts.cpt['0|' + causeVal] = 0.97;
      cfts.cpt['1|' + causeVal] = 0.02;
      cfts.cpt['2|' + causeVal] = 0.01;
    }
  }

  var ffox = new bnNode();
  ffox.bnet = bnet;
  bnet['ffox'] = ffox;
  ffox.name = 'ffox';
  ffox.range = [0, 14, 30];
  ffox.parents = ['cause'];
  ffox.cpt = {};

  for (var i=0; i<cause.range.length; i++) {
    var causeVal = cause.range[i];
    if (causeVal[1] == 'D') {
      // dns failure
      ffox.cpt['0|' + causeVal] = 0.1;
      ffox.cpt['14|' + causeVal] = 0.3;
      ffox.cpt['30|' + causeVal] = 0.9;
    } else {
      // no dns failure
      ffox.cpt['0|' + causeVal] = 0.97;
      ffox.cpt['14|' + causeVal] = 0.01;
      ffox.cpt['30|' + causeVal] = 0.02;
    }
  }

  rv += "<br><h2>Belief propagation test</h2>";
  for (var childId=0; childId<cause.children.length; childId++) {
    var childName = cause.children[childId];
    bnet[childName].calcLambda();
    bnet[childName].calcLambdaMessages();
  }

  cause.updateBelief();
  
  for (var childId=0; childId<cause.children.length; childId++) {
    var childName = cause.children[childId];
    bnet[childName].updateBelief();
  }

  for (var nodeName in bnet) {
    var node = bnet[nodeName];
    for (var valId=0; valId<node.range.length; valId++) {
      var val = node.range[valId];
      rv += nodeName + '.bel[' + val + '] == ' + Math.round(node.bel[val] * 10000) / 10000 + '<br>';
    }
  }

  rv += "<h2>Evidence test</h2>";

  obconn.setEvidence('FAIL');

  for (var childId=0; childId<cause.children.length; childId++) {
    var childName = cause.children[childId];
    bnet[childName].calcLambda();
    bnet[childName].calcLambdaMessages();
  }

  cause.updateBelief();
  
  for (var childId=0; childId<cause.children.length; childId++) {
    var childName = cause.children[childId];
    bnet[childName].updateBelief();
  }

  for (var nodeName in bnet) {
    var node = bnet[nodeName];
    for (var valId=0; valId<node.range.length; valId++) {
      var val = node.range[valId];
      rv += nodeName + '.bel[' + val + '] == ' + Math.round(node.bel[val] * 10000) / 10000 + '<br>';
    }
  }

  return rv;
}

// A --> B
function bnTest() {
  var errors = "";

  function assertEquals(x, y) {
    if (!eval("(" + x + ") == (" + y + ")")) {
      errors += "assertion failed: ";
      errors += x + " == " + eval(x) + ", ";
      errors += y + " == " + eval(y) + "<br>";
    }
  }

  var bnet = {};

  var a = new bnNode();
  a.bnet = bnet;
  a.name = "A";
  a.cpt = {};
  a.range = ['a1', 'a2'];
  a.children = ["B"];
  a.cpt["a1|"] = 0.1;
  a.cpt["a2|"] = 0.9;

  var b = new bnNode();
  b.bnet = bnet;
  b.name = "B";
  b.cpt = {};
  b.range = ['b1', 'b2'];
  b.parents = ["A"];
  b.cpt["b1|a1"] = 0.3;
  b.cpt["b1|a2"] = 0.8;
  b.cpt["b2|a1"] = 0.7;
  b.cpt["b2|a2"] = 0.2;

  bnet["A"] = a;
  bnet["B"] = b;

  b.calcLambda();
  b.calcLambdaMessages();
  
  a.updateBelief();
  b.updateBelief();

  assertEquals("a.pi['a1']", "0.1");
  assertEquals("a.pi['a2']", "0.9");

  assertEquals("a.piMessages[0]['a1']", "0.1");
  assertEquals("a.piMessages[0]['a2']", "0.9");

  assertEquals("a.lambdaFromChildren['B']['a1']", "1");
  assertEquals("a.lambdaFromChildren['B']['a2']", "1");

  assertEquals('b.lambda["b1"]', '1');
  assertEquals('b.lambda["b2"]', '1');

  assertEquals('Math.round(b.pi["b1"] * 10000)', '.75 * 10000');
  assertEquals('Math.round(b.pi["b2"] * 10000)', '.25 * 10000');

  assertEquals('b.lambdaMessages[0]["a1"]', '1');
  assertEquals('b.lambdaMessages[0]["a2"]', '1');

  assertEquals("b.piFromParents['A']['a1']", "0.1");
  assertEquals("b.piFromParents['A']['a2']", "0.9");

  assertEquals('Math.round(a.bel["a1"] * 10000)', '.1 * 10000');
  assertEquals('Math.round(a.bel["a2"] * 10000)', '.9 * 10000');

  assertEquals('Math.round(b.bel["b1"] * 10000)', '.75 * 10000');
  assertEquals('Math.round(b.bel["b2"] * 10000)', '.25 * 10000');

  b.setEvidence('b1');

  assertEquals('b.lambda["b1"]', '1');
  assertEquals('b.lambda["b2"]', '0');

  b.calcLambdaMessages();
  
  a.updateBelief();
  b.updateBelief();

  assertEquals('Math.round(a.bel["a1"] * 10000)', '.04 * 10000');
  assertEquals('Math.round(a.bel["a2"] * 10000)', '.96 * 10000');

  assertEquals('Math.round(b.bel["b1"] * 10000)', '1.0 * 10000');
  assertEquals('Math.round(b.bel["b2"] * 10000)', '0.0 * 10000');
  
  assertEquals("a.lambdaFromChildren['B']['a1']", "0.3");
  assertEquals("a.lambdaFromChildren['B']['a2']", "0.8");

  b.clearEvidence();

  // add evidence for a1 through a dummy node
  var c = new bnNode();
  c.bnet = bnet;
  c.name = "C";
  c.cpt = {};
  c.range = ['c0'];
  c.parents = ["A"];
  c.cpt["c0|a1"] = 1.0;
  c.cpt["c0|a2"] = 0.0;
  bnet["C"] = c;
  a.children.push("C");

  b.calcLambda();
  b.calcLambdaMessages();

  c.setEvidence("c0");
  c.calcLambda();
  c.calcLambdaMessages();
  a.updateBelief();
  b.updateBelief();
  c.updateBelief();

  assertEquals("a.lambdaFromChildren['C']['a1']", "1.0");
  assertEquals("a.lambdaFromChildren['C']['a2']", "0.0");

  assertEquals('Math.round(a.bel["a1"] * 10000)', '1.0 * 10000');
  assertEquals('Math.round(a.bel["a2"] * 10000)', '0.0 * 10000');

  assertEquals('Math.round(c.bel["c0"] * 10000)', '1.0 * 10000');

  assertEquals("a.piMessages[0]['a1']", "1.0");
  assertEquals("a.piMessages[0]['a2']", "0.0");

  assertEquals("b.piFromParents['A']['a1']", "1.0");
  assertEquals("b.piFromParents['A']['a2']", "0.0");

  assertEquals('Math.round(b.bel["b1"] * 10000)', '0.3 * 10000');
  assertEquals('Math.round(b.bel["b2"] * 10000)', '0.7 * 10000');

  if (errors == "") {
    errors = "All tests passed.";
  }

  return errors;
};

function bnNode() {
  this.cpt = {},                // cpt["var|parents"]
  this.bel = {};                // belief
  this.lambda = {};		// lambda calculation used locally	    
  this.pi = {};			// pi calculation used locally	    
  this.lambdaFromChildren = {};	// lambda messages receive from children 
  this.piFromParents = {};	// pi messages received from parents      
  this.range = [];		// list of possible values for this variable
  this.children = [];           // list of children
  this.parents = [];            // list of parents
  this.lambdaMessages = [];     // lambda messages to be sent to parents
  this.piMessages = [];         // pi messages to be sent to children
};

bnNode.prototype = {
  bnet: null,             // associative array of all nodes
  name: "",               // name of this variable
  evidence: false,        // whether this is an evidence node

  setEvidence: function(observedValue) {
    for (var i=0; i<this.range.length; i++) {
      var val = this.range[i];
      if (val == observedValue) {
	this.lambda[val] = 1;
      } else {
	this.lambda[val] = 0;
      }
    }
    this.evidence = true;
  },

  clearEvidence: function() {
    this.evidence = false;
  },

//   activate: function(fromNode) {
//     // recursively request messages from all neighbors, and then send
//     // a message back.  Like a depth first search.
//     //
//     // 1. For each neighbor except fromNode, check to see if I have a
//     // message from them.
//     // 2. Recursively activate all nodes whom I haven't heard from.
//     // 3. Send a message to all other fromNode.

//     var fromNodeName = fromNode.name;

//     for (childId=0; childId<this.children.length; childId++) {
//       var childName = this.children[childId];
//       var childNode = this.bnet[childName];
//       if (childName != fromNodeName &&
// 	  lambdaFromChildren[childName] == undefined) {
// 	childNode.activate(this);
//       }
//     }

//     for (parentId=0; parentId<this.parents.length; parentId++) {
//       var parentName = this.parents[parentId];
//       var parentNode = this.bnet[parentName];
//       if (parentName != fromNodeName &&
// 	  piFromParents[parentName] == undefined) {
// 	childNode.activate(this);
//       }
//     }
//   },

//   sendMessages: function() {
//     // determine if this node can send any messages; if so, send them.
//     // Return whether it was able to send any messages.
//     //
//     // For each neighbor, check to see if we have received messages
//     // from all other neighbors.  If so, send a message to the
//     // neighbor.
//     //
//     // Next check to see if that neighbor can now send any new
//     // messages to its neighbors.

//     // If we are an edge node with only one neighbor, then we can
//     // propagate a message before receiving any other messages.

//     if (this.parents.length == 1 && this.children.length == 0) {
//       // Leaf node.  Calculate lambda and send lambda message to
//       // parent.
//       this.calcLambda();
//       this.calcLambdaMessages();
//     } else if (this.parents.length == 0 && this.children.length == 1) {
//       // Root node.  calculate pi and send pi message to child.
//       this.calcPi();
//       this.calcPiMessages();
//     }
    
//     // Otherwise, if we've received lambda messages from all our
//     // children and pi messages from all our parents, then we can send
//     // lambda messages to our parents and pi messages to our children.
//     if (this.lambdaFromChildren.length == children.length) {
//       this.calcLambda();
//       this.calcLambdaMessages();
//     }

//     if (this.piFromParents.length == parents.length) {
//       this.calcPi();
//       this.calcPiMessages();
//     }
//   },

  updateBelief: function() {
    // update beliefs
    
    // sum of all beliefs for normalization purposes
    var belSum = 0.0;

    // calculate pi and lambda values
    this.calcLambda();
    this.calcPi();

    // update belief
    for (var i=0; i<this.range.length; i++) {
      var val = this.range[i];
      this.bel[val] = this.lambda[val] * this.pi[val];
      belSum += this.bel[val];
    }
    
    // normalize beliefs
    for (var i=0; i<this.range.length; i++) {
      var val = this.range[i];
      this.bel[val] /= belSum;
    }

    this.calcLambdaMessages();
    this.calcPiMessages();
  },

  nextParentValueIds: function(parentValueIds) {
    // return the next possible combination of values for the parents
    // in parentValues.  parentValues is an array of parent value
    // indices in the same order as this.parents.  If parentValues is
    // null, start with the first possible combination of parent
    // values.
    //
    // The first parent value changes most rapidly.  If there are no
    // more values, return null.

    var nextValues = new Array();
    
    if (parentValueIds == null) {
      for (var i=0; i<this.parents.length; i++) {
	nextValues.push(0);
      }
    } else {
      var noMoreValues = true;
      for (var i=0; i<this.parents.length; i++) {
	var parent = this.parents[i];
	var maxValue = this.bnet[parent].range.length - 1;
	if (parentValueIds[i] < maxValue) {
	  parentValueIds[i]++;
	  noMoreValues = false;
	  break;
	} else {
	  parentValueIds[i] = 0;
	}
      }
      
      if (noMoreValues) {
	nextValues = null;
      } else {
	nextValues = parentValueIds;
      }
    }

    return nextValues;
  },

  getParentValues: function(parentValueIds) {
    // given a list of parent value ids, return the actual parent values.
    //
    var parentValues = new Array();

    for (var parentId=0; parentId<this.parents.length; parentId++) {
      var parentName = this.parents[parentId];
      var parentNode = this.bnet[parentName];
      var parentValueId = parentValueIds[parentId];
      var parentValue = parentNode.range[parentValueId];
      parentValues.push(parentValue);
    }
    return parentValues;
  },

  calcLambda: function() {
    if (this.evidence) {
      // if we already have evidence, no need to update lambda
      return;
    }
    // calculate lambda from messages received from children
    for (var i=0; i<this.range.length; i++) {
      var val = this.range[i];
      this.lambda[val] = 1.0;
      for (var j=0; j<this.children.length; j++) {
	var child = this.children[j];
	if (this.lambdaFromChildren[child] == undefined) {
	  this.agent.dump(this.name + ".calcLambda(): lambdaFromChildren[" + child + "] not defined");
	}
	this.lambda[val] *= this.lambdaFromChildren[child][val];
      }
    }
  },
  
  calcPi: function() {
    // calculate pi values
    
    
    for (var i=0; i<this.range.length; i++) {
      var val = this.range[i];
      this.pi[val] = 0;
      
      var parentValueIds = null;
      while (true) {
	var parentValueIds = this.nextParentValueIds(parentValueIds);
	if (parentValueIds == null) {
	  break;
	}
	var parentValues = this.getParentValues(parentValueIds);
	
	var prod = 1;
	for (var j=0; j<this.parents.length; j++) {
	  var parent = this.parents[j];
	  var piFromParent = this.piFromParents[parent][parentValues[j]];
	  prod *= piFromParent;
	}
	var parentValues = this.getParentValues(parentValueIds);

	var key = val + "|" + parentValues.join(" ");
	var parentsProb = this.cpt[key];
	if (parentsProb == null) {
	  this.agent.dump("calcPi: missing CPT: parentsProb[" + key + "]");
	}

	prod *= parentsProb;
	this.pi[val] += prod;
      }
    }
  },

  calcLambdaMessages: function() {
    var lambdaMessages = new Array();
    for (var toParentId=0; toParentId<this.parents.length; toParentId++) {
      var lambdaMessage = this.calcLambdaMessage(toParentId);
      lambdaMessages.push(lambdaMessage);
    }
    this.lambdaMessages = lambdaMessages;
  },

  calcLambdaMessage: function(toParentId) {
    // calculate the lambda message to send to toParentId
    // for each parent to whom to send a message
    var lambdaMessage = {};
    var toParent = this.parents[toParentId];
    var toParentNode = this.bnet[toParent];

    // for each parent value
    for (var toParentValId=0;
	 toParentValId<toParentNode.range.length;
	 toParentValId++) {
      var toParentVal = toParentNode.range[toParentValId];
	
      // sum over x of lambda(x) * innerSum
      var outerSum = 0;

      for (var valId = 0; valId < this.range.length; valId++) {
	var val = this.range[valId];

	// sum of conditional probabilities from parents
	var innerSum = 0;
	var parentValueIds = null;
	while (true) {
	  var parentValueIds = this.nextParentValueIds(parentValueIds);
	  if (parentValueIds == null) {
	    break;
	  }
	  // u_k: k!=i
	  if (parentValueIds[toParentId] != toParentValId) {
	    continue;
	  }

	  var parentValues = this.getParentValues(parentValueIds);

	  // compute the product of all pi messages received from
	  // other parents.
	  var prod = 1.0;
	  for (var fromParentId=0;
	       fromParentId<this.parents.length;
	       fromParentId++) {
	    var fromParent = this.parents[fromParentId];
	    if (fromParentId == toParentId) {
	      continue;
	    }
	    var fromParentValue = parentValues[fromParentId];
	    prod *= this.piFromParents[fromParent][fromParentValue];
	  }

	  var key = val + "|" + parentValues.join(",");
	  var parentsProb = this.cpt[key];
	  if (parentsProb == null) {
	    this.agent.dump("calcLambdaMessage: missing CPT: parentsProb[" +
			    key + "]");
	  }
	  prod *= parentsProb;
	  innerSum += prod;
	}
	outerSum += this.lambda[val] * innerSum;
      }
      lambdaMessage[toParentVal] = outerSum;
    }
    toParentNode.lambdaFromChildren[this.name] = lambdaMessage;
    return lambdaMessage;
  },

  calcPiMessages: function() {
    var piMessages = new Array();

    for (var childId = 0; childId<this.children.length; childId++) {
      var piMessage = this.calcPiMessage(childId);
      piMessages.push(piMessage);
    }
    this.piMessages = piMessages;
  },

  calcPiMessage: function(toChildId) {
    // calculate the pi messages that go to children.
    //
    var debug = false;
    var piMessage = {};
    var toChild = this.children[toChildId];
    var toChildNode = this.bnet[toChild];
    var sum = 0; // for normalization
    for (var i=0; i<this.range.length; i++) {
      var val = this.range[i];

      piMessage[val] = this.pi[val];
      
      for (var otherChildId = 0; otherChildId<this.children.length; otherChildId++) {
	var otherChild = this.children[otherChildId];
	if (toChildId == otherChildId) {
	  continue;
	}
	piMessage[val] *= this.lambdaFromChildren[otherChild][val];
      }
      sum += piMessage[val];
    }

    for (var i=0; i<this.range.length; i++) {
      var val = this.range[i];
      piMessage[val] /= sum;
    }
    toChildNode.piFromParents[this.name] = piMessage;

    return piMessage;
  }
};
