/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is ServStats.
 *
 * The Initial Developer of the Original Code is
 * George J. Lee <gjl@csail.mit.edu>
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

var diagTool;

//////////////////////
// DiagTool Object
function DiagTool() {};
DiagTool.prototype = {
  agent: null,
  debug: false,

  dump: function(aMessage) {
    if (this.debug) {
      var consoleService = Components.classes["@mozilla.org/consoleservice;1"].getService(Components.interfaces.nsIConsoleService);
      consoleService.logStringMessage("DiagTool: " + aMessage);
    }
  },

  start: function() {
    this.dump("Starting DiagTool");
    var prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);

    var observerService = Components.classes['@mozilla.org/observer-service;1'].getService(Components.interfaces.nsIObserverService);
    observerService.addObserver(this, 'diagtool-set-last-failure', false);
    observerService.addObserver(this, 'diagtool-set-diag-state', false);
    observerService.addObserver(this, 'diagtool-set-diag-result', false);
    observerService.addObserver(this, 'diagtool-servstats-disabled', false);

    this.agent = Components.classes["@capri.csail.mit.edu/ua;1"].getService(Components.interfaces.nsIDiagAgent);

    gBrowser.addProgressListener(this.agent,
				 Components.interfaces.nsIWebProgress.NOTIFY_STATE_NETWORK |
				 Components.interfaces.nsIWebProgress.NOTIFY_STATE_DOCUMENT);

    if (this.debug) {
      // add a "run tests" menu item
      var menu = document.getElementById("diag-popup");
      var runTestsItem = document.createElement("menuitem");
      runTestsItem.setAttribute("label", "Run Tests");
      runTestsItem.setAttribute("oncommand", "diagTool.runTests()");
      menu.appendChild(runTestsItem);
    }
  },

  stop: function() {
    try {
      var observerService = Components.classes['@mozilla.org/observer-service;1'].getService(Components.interfaces.nsIObserverService);
      observerService.removeObserver(this, "diagtool-set-last-failure");
      observerService.removeObserver(this, "diagtool-set-diag-state");
      observerService.removeObserver(this, "diagtool-set-diag-result");
      observerService.removeObserver(this, 'diagtool-servstats-disabled');
      observerService = null;

      gBrowser.removeProgressListener(this.agent);
    } catch (e) {};
  },

  probeHosts: function() {
    hosts = ['localhost'];
    this.agent.probeHosts(hosts, hosts.length);
  },

  getHist: function() {
    return this.connHist;
  },

  setLastFailure: function(host) {
    // Set the "Last Failure: " label in the toolbar.  This gets
    // called when a new failure occurs.
    
    var lastFailureLabel = document.getElementById("diag-last-failure");
    var value = host;
    lastFailureLabel.setAttribute("value", value);

    // Don't allow users to diagnose localhost
    var disabled = (host == 'localhost');
    var diagButton = document.getElementById("diag-button");
    diagButton.setAttribute("disabled", disabled);

    // display diagnostic info
    var toolbar = document.getElementById("diag-toolbar");
    toolbar.setAttribute("hidden", "false");
    
    var collapsedItems = ['diag-last-failure-item',
			  'diag-button',
			  'diag-sep-1',
			  'diag-result-item',
			  'diag-sep-2'];

    for (var itemIdx=0; itemIdx<collapsedItems.length; itemIdx++) {
      var collapsedItem = document.getElementById(collapsedItems[itemIdx]);
      collapsedItem.setAttribute("collapsed", "false");
    }
    
    this.setDiagState('|true');
    this.setDiagResult('');
  },

  setDiagState: function(data) {
    // Set the diagnosis progress indicator given a state value and a
    // diagComplete boolean separated by a '|'
    var parts = data.split('|');
    var state = parts[0];
    var diagComplete = (parts[1] == 'true');
    
    var lastFailureLabel = document.getElementById("diag-state");
    var value = state;
    lastFailureLabel.setAttribute("value", value);

    // If the diagnosis is not yet complete, gray out the current
    // result and the why button and activate the throbber.
    var diagResultLabel = document.getElementById("diag-result");
    var whyButton = document.getElementById("diag-button");
    var diagThrobber = document.getElementById("diag-throbber");
    diagResultLabel.setAttribute("disabled", !diagComplete);
    whyButton.setAttribute("disabled", !diagComplete);
    diagThrobber.setAttribute("busy", !diagComplete);
  },

  hideToolbar: function() {
    var toolbar = document.getElementById("diag-toolbar");
    toolbar.setAttribute("hidden", "true");
  },

  setDiagResult: function(result) {
    // Set the "Diagnosis: " label in the toolbar and set the
    // appropriate tooltip.
    //
    // remove all children from tooltipBox
    var tooltipBox = document.getElementById("diag-toolbar-diagnosis-tip-box");
    while (tooltipBox.firstChild) {
      tooltipBox.removeChild(tooltipBox.firstChild);
    }

    var diagResultItem = document.getElementById("diag-result-item");
    diagResultItem.setAttribute("collapsed", "false");

    var parts = result.split('|');
    var diagResult = parts[0];
    var diagResultLabel = document.getElementById("diag-result");
    diagResultLabel.setAttribute("value", diagResult);
    diagResultLabel.setAttribute("disabled", "false");
    diagResultLabel.removeAttribute("tooltip");

    // set the tooltip if there's more than one part
    if (parts.length > 1) {
      var tooltiptext = null;
      tooltiptext = parts[1];
      if (tooltiptext != null) {
	tooltiptext = tooltiptext.replace(/,/, ': ');
	var tooltipRows = tooltiptext.split('\n');
	for (var row=0; row<tooltipRows.length; row++) {
	  var descNode = document.createElement("description");
	  descNode.textContent = tooltipRows[row];
	  tooltipBox.appendChild(descNode);
	}
	diagResultLabel.setAttribute("tooltip", "diag-toolbar-diagnosis-tip");
      }
    }
  },
  
  setServStatsDisabled: function(disabled) {
    var button = document.getElementById("diag-stats-button");
    if (disabled != 'true') {
      disabled = 'false';
    }
    button.setAttribute("disabled", disabled);
  },
  
  observe: function(subject, topic, data) {
    switch (topic) {
      case "diagtool-set-last-failure":
        this.setLastFailure(data);
        break;
      case "diagtool-set-diag-state":
        this.setDiagState(data);
        break;
      case "diagtool-set-diag-result":
        this.setDiagResult(data);
        break;
      case "diagtool-servstats-disabled":
        this.setServStatsDisabled(data);
        break;
    }

  },

  runTests: function() {
    // run tests
    this.agent.runTests();
  },

  openHistory: function() {
    window.openDialog('chrome://diag/content/history.xul',
		      'history',
		      'chrome,width=200,height=100');
  },

  openOptions: function() {
    window.openDialog('chrome://diag/content/options.xul',
		      'options',
		      'chrome');
  },

  requestServStats: function() {
    this.dump("Requesting stats");
    this.agent.requestServStats(gBrowser.currentURI.asciiHost);
  },

  requestDiagnosis: function() {
    this.dump("Requesting diagnosis");
    this.agent.requestDiagnosis(null);
  }
};

function dtLoad()
{
  diagTool = new DiagTool();
  diagTool.start();
};

function dtUnload()
{
  diagTool.stop();
  diagTool = null;
};

window.addEventListener("load", dtLoad, false);
window.addEventListener("unload", dtUnload, false);

